# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import hashlib
import tempfile

from unittest import TestCase

from mock import patch, call, Mock

from managesf.tests import dummy_conf
from managesf.model.yamlbkd.resources.gitrepository import GitRepositoryOps


class GitRepositoryOpsTest(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()

    def test_create(self):
        o = GitRepositoryOps(self.conf, None)

        kwargs = {'name': 'space/g1',
                  'description': 'A description',
                  'acl': 'a1'}

        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.create_project') as cp, \
            patch('managesf.services.gerrit.utils.GerritRepo.clone'), \
            patch.object(GitRepositoryOps,
                         'install_acl') as ia, \
            patch.object(GitRepositoryOps,
                         'create_branches') as cb, \
            patch.object(GitRepositoryOps,
                         'install_git_review_file') as ig:
            ia.return_value = []
            cb.return_value = []
            ig.return_value = []
            logs = o.create(**kwargs)
            self.assertEqual(len(cp.call_args_list), 1)
            self.assertEqual(cp.call_args_list[0],
                             call('space/g1', 'A description',
                                  ['Administrators']))
            self.assertEqual(len(logs), 0)

        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.create_project') as cp, \
            patch('managesf.services.gerrit.utils.GerritRepo.clone'), \
            patch.object(GitRepositoryOps,
                         'install_acl') as ia, \
            patch.object(GitRepositoryOps,
                         'create_branches') as cb, \
            patch.object(GitRepositoryOps,
                         'install_git_review_file') as ig:
            ia.return_value = []
            cb.return_value = []
            ig.return_value = []
            cp.side_effect = Exception('Random Error')
            logs = o.create(**kwargs)
            self.assertEqual(len(logs), 1)
            self.assertIn('Repo create: err API returned Random Error',
                          logs)

    def test_install_git_review_file(self):
        o = GitRepositoryOps(self.conf, {})

        MGR = Mock()

        logs = o.install_git_review_file(MGR, 'space/g1', 'br1')
        self.assertTrue(MGR.push_branch.called)
        self.assertEqual(
            MGR.push_branch.call_args,
            call('br1',
                 {'.gitreview': '[gerrit]\nhost=tests.dom\nport=2929'
                  '\nproject=space/g1\ndefaultbranch=br1\n'})
        )
        self.assertEqual(len(logs), 0)

    def test_create_branches(self):
        o = GitRepositoryOps(self.conf, {})
        o._set_client()

        MGR = Mock()
        MGR.list_remote_branches.return_value = {
            'HEAD': 'master',
            'master': '100',
            'dev3': '125'}

        kwargs = {'name': 'space/g1',
                  'default-branch': '',
                  'branches': {}}
        logs = o.create_branches(MGR, **kwargs)
        self.assertTrue(MGR.list_remote_branches.called)
        self.assertTrue(not MGR.create_remote_branch.called)
        self.assertTrue(not MGR.delete_remote_branch.called)
        self.assertEqual(len(logs), 0)

        MGR.reset_mock()

        MGR.list_remote_branches.return_value = {
            'HEAD': 'rpm-master',
            'master': '100',
            'dev3': '125'}
        kwargs = {'name': 'space/g1',
                  'default-branch': '',
                  'branches': {
                      'dev': '123',
                      'dev2': '124',
                      'dev3': '0'}}
        with patch.object(GitRepositoryOps,
                          'set_default_branch') as sdb, \
            patch.object(GitRepositoryOps,
                         'install_git_review_file') as igrf:
            logs = o.create_branches(MGR, **kwargs)
        MGR.list_remote_branches.assert_called()
        MGR.create_remote_branch.assert_called()
        igrf.assert_called()
        sdb.assert_not_called()
        MGR.delete_remote_branch.assert_called()
        self.assertEqual(len(logs), 0)
        self.assertIn(call('dev2', '124'),
                      MGR.create_remote_branch.call_args_list)
        self.assertIn(call('dev', '123'),
                      MGR.create_remote_branch.call_args_list)
        self.assertIn(call(MGR, 'space/g1', 'dev2'), igrf.call_args_list)
        self.assertIn(call(MGR, 'space/g1', 'dev'), igrf.call_args_list)
        self.assertTrue(len(igrf.call_args_list), 2)
        self.assertTrue(len(MGR.create_remote_branch.call_args_list), 2)
        self.assertIn(call('dev3'), MGR.delete_remote_branch.call_args_list)
        self.assertTrue(len(MGR.delete_remote_branch.call_args_list), 1)

        MGR.reset_mock()

        MGR.list_remote_branches.return_value = {
            'HEAD': 'master',
            'master': '100',
            'dev3': '125'}
        kwargs = {'name': 'space/g1',
                  'default-branch': 'dev',
                  'branches': {
                      'dev': '123'}}
        with patch.object(GitRepositoryOps,
                          'set_default_branch') as sdb, \
            patch.object(GitRepositoryOps,
                         'install_git_review_file') as igrf:
            logs = o.create_branches(MGR, **kwargs)

        MGR.list_remote_branches.assert_called()
        MGR.create_remote_branch.assert_called()
        sdb.assert_called()
        igrf.assert_called()
        MGR.delete_remote_branch.assert_not_called()
        self.assertEqual(len(logs), 0)
        self.assertIn(call('dev', '123'),
                      MGR.create_remote_branch.call_args_list)
        self.assertTrue(len(MGR.create_remote_branch.call_args_list), 1)
        self.assertIn(call(MGR, 'space/g1', 'dev'), igrf.call_args_list)
        self.assertTrue(len(igrf.call_args_list), 1)
        self.assertIn(call('space/g1', 'dev'), sdb.call_args_list)
        self.assertTrue(len(sdb.call_args_list), 1)

        MGR.reset_mock()

        MGR.list_remote_branches.return_value = {
            'HEAD': 'master',
            'master': '100'}
        kwargs = {'name': 'space/g1',
                  'default-branch': 'rpm-master',
                  'branches': {}}
        with patch.object(GitRepositoryOps,
                          'set_default_branch') as sdb, \
            patch.object(GitRepositoryOps,
                         'install_git_review_file') as igrf:

            logs = o.create_branches(MGR, **kwargs)

        MGR.list_remote_branches.assert_called()
        MGR.create_remote_branch.assert_called()
        sdb.assert_called()
        igrf.assert_called()
        not MGR.delete_remote_branch.assert_not_called()
        self.assertEqual(len(logs), 0)
        self.assertIn(call('rpm-master', 'HEAD'),
                      MGR.create_remote_branch.call_args_list)
        self.assertTrue(len(MGR.create_remote_branch.call_args_list), 1)
        self.assertIn(call(MGR, 'space/g1', 'rpm-master'), igrf.call_args_list)
        self.assertTrue(len(igrf.call_args_list), 1)
        self.assertIn(call('space/g1', 'rpm-master'), sdb.call_args_list)
        self.assertTrue(len(sdb.call_args_list), 1)

    def test_install_acl(self):
        new = {
            'resources': {
                'acls': {
                    'a1': {
                        'file': '',
                        'groups': ['g1'],
                    }
                },
                'groups': {
                    'g1': {
                        'name': 'sf/g1',
                        'members': ['body@sftests.com'],
                    }
                }
            }
        }

        o = GitRepositoryOps(self.conf, new)
        o._set_client()

        kwargs = {'name': 'space/g1',
                  'description': 'A description',
                  'acl': ''}

        db = {'Administrators': '666',
              'Anonymous Users': '777',
              'Service Users': '888'}

        MGR = Mock()

        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.get_group_id') as ggi:
            ggi.side_effect = lambda x: db[x]
            logs = o.install_acl(MGR, **kwargs)
        self.assertIn(call('Administrators'), ggi.call_args_list)
        self.assertIn(call('Anonymous Users'), ggi.call_args_list)
        self.assertIn(call('Service Users'), ggi.call_args_list)
        self.assertEqual(len(ggi.call_args_list), 3)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "666\\tAdministrators")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "777\\tAnonymous Users")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "888\\tService Users")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "description = A description")), 0)
        self.assertEqual(len(logs), 0)

        kwargs = {'name': 'space/g1',
                  'description': 'A description',
                  'acl': 'a1'}

        db = {'Administrators': '666',
              'Anonymous Users': '777',
              'Service Users': '888',
              'sf/g1': 999}

        MGR.reset_mock()

        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.get_group_id') as ggi:
            ggi.side_effect = lambda x: db[x]
            logs = o.install_acl(MGR, **kwargs)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "666\\tAdministrators")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "777\\tAnonymous Users")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "888\\tService Users")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "999\\tsf/g1")), 0)
        self.assertGreater(
            int(str(MGR.push_config.call_args).find(
                "description = A description")), 0)
        self.assertEqual(len(logs), 0)

        MGR.reset_mock()

        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.get_group_id') as ggi:
            ggi.side_effect = lambda x: db[x]
            MGR.push_config.side_effect = Exception('Random error')
            logs = o.install_acl(MGR, **kwargs)
        self.assertListEqual(['Random error'], logs)

    def test_update(self):
        with patch.object(GitRepositoryOps, 'install_acl') as ia, \
                patch('managesf.services.gerrit.utils.GerritRepo.clone'), \
                patch.object(GitRepositoryOps, 'create_branches') as cb:
            ia.return_value = ['log']
            cb.return_value = []
            o = GitRepositoryOps(self.conf, None)
            logs = o.update(name='space/p1', k='v')
            self.assertTrue(ia.called)
            self.assertTrue(cb.called)
            self.assertIn('log', logs)

    def test_delete(self):
        o = GitRepositoryOps(self.conf, None)

        kwargs = {'name': 'space/g1'}
        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.delete_project') as dp:
            logs = o.delete(**kwargs)
            self.assertEqual(len(dp.call_args_list), 1)
            self.assertEqual(dp.call_args_list[0],
                             call('space/g1', True,))
            self.assertEqual(len(logs), 0)

    def test_get_all(self):
        a1 = """[project]
    description = This is the project p1
[access "refs/*"]
  read = group Anonymous Users
    read = group p1-core
  owner = group p1-ptl
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group p1-core
    label-Verified = -2..+2 group p1-ptl
    label-Workflow = -1..+1 group p1-core
    submit = group p1-ptl
    read = group Anonymous Users
    read = group p1-core
[access "refs/meta/config"]
    read = group p1-core
[receive]
    requireChangeId = true
[submit]
    mergeContent = false
    action = fast forward only
"""

        a2 = """[submit]
    mergeContent = false
    action = fast forward only
"""

        clean_a1 = """[access "refs/*"]
  read = group Anonymous Users
    read = group p1-core
  owner = group p1-ptl
[access "refs/heads/*"]
    label-Code-Review = -2..+2 group p1-core
    label-Verified = -2..+2 group p1-ptl
    label-Workflow = -1..+1 group p1-core
    submit = group p1-ptl
    read = group Anonymous Users
    read = group p1-core
[access "refs/meta/config"]
    read = group p1-core
[receive]
    requireChangeId = true
[submit]
    mergeContent = false
    action = fast forward only
"""

        m = hashlib.md5()
        m.update(clean_a1.encode())
        a1_id = m.hexdigest()
        m = hashlib.md5()
        m.update(a2.encode())
        a2_id = m.hexdigest()

        def fake_get_projects():
            return ['p1', 'p2']

        def fake_repo_utils(name, conf):
            class FakeGerritRepo():
                def __init__(self, name, conf):
                    self.name = name
                    self.conf = conf

                def get_raw_acls(self):
                    data = {
                        'p1': a1,
                        'p2': a2,
                    }
                    f, p = tempfile.mkstemp()
                    os.close(f)
                    open(p, 'w').write(data[self.name])
                    return p
            return FakeGerritRepo(name, conf)

        o = GitRepositoryOps(self.conf, None)
        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.get_projects') as gps, \
                patch('managesf.services.gerrit.utils.GerritRepo') as gr:
            gps.side_effect = fake_get_projects
            gr.side_effect = fake_repo_utils
            logs, tree = o.get_all()
            self.assertIn('repos', tree.keys())
            self.assertIn('acls', tree.keys())
            self.assertIn(a1_id, tree['acls'].keys())
            self.assertIn(a2_id, tree['acls'].keys())
            self.assertEqual(len(tree['acls'].keys()), 2)
            self.assertIn('p1', tree['repos'].keys())
            self.assertIn('p2', tree['repos'].keys())
            self.assertEqual(len(tree['repos'].keys()), 2)
            self.assertDictEqual(tree['repos']['p1'],
                                 {'name': 'p1',
                                  'description': 'This is the project p1',
                                  'acl': a1_id})
            self.assertDictEqual(tree['repos']['p2'],
                                 {'name': 'p2',
                                  'acl': a2_id})
            self.assertEqual(tree['acls'][a1_id]['file'], clean_a1)
            self.assertIn('p1-ptl', tree['acls'][a1_id]['groups'])
            self.assertIn('p1-core', tree['acls'][a1_id]['groups'])
            self.assertEqual(len(tree['acls'][a1_id]['groups']), 2)
            self.assertEqual(tree['acls'][a2_id]['file'], a2)
            self.assertEqual(len(tree['acls'][a2_id]['groups']), 0)



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from unittest import TestCase
from webtest import TestApp
from pecan import load_app
from mock import patch

from managesf.tests import dummy_conf


class V2FunctionalTest(TestCase):
    def setUp(self):
        c = dummy_conf()
        self.config = {'services': c.services,
                       'gerrit': c.gerrit,
                       'app': c.app,
                       'admin': c.admin,
                       'sqlalchemy': c.sqlalchemy,
                       'managesf': c.managesf,
                       'policy': c.policy,
                       'resources': c.resources,
                       'api': c.api, }
        # App must be loaded before we can import v2 managers
        self.app = TestApp(load_app(self.config))

    def tearDown(self):
        # Remove the sqlite db
        os.unlink(self.config['sqlalchemy']['url'][len('sqlite:///'):])


class TestManageSFV2ResourcesController(V2FunctionalTest):

    @patch('managesf.controllers.api.v2.resources.manager')
    def test_get_resources(self, rm):
        with patch('managesf.controllers.api.v2.base.authorize'):
            environ = {'REMOTE_USER': 'user'}
            # not the real format of the answer here but who cares, it's a test
            retval = {'skipped': 0,
                      'total': 10,
                      'limit': 1,
                      'results': ['yo yo']}
            rm.resources.get.return_value = retval
            response = self.app.get('/v2/resources/',
                                    extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200, response.text)
            self.assertEqual('yo yo', response.json['results'][0])
            # check specific args
            response = self.app.get('/v2/resources/'
                                    '?get_missing_resources=true',
                                    extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200, response.text)
            _, kwargs = rm.resources.get.call_args
            self.assertEqual(True,
                             kwargs.get('get_missing_resources'),
                             kwargs)

    @patch('managesf.controllers.api.v2.resources.manager')
    def test_validate_resources(self, rm):
        with patch('managesf.controllers.api.v2.base.authorize'):
            environ = {'REMOTE_USER': 'user'}
            # not the real format of the answer here but who cares, it's a test
            retval = True, 'yo yo'
            rm.resources.create.return_value = retval
            response = self.app.post('/v2/resources/',
                                     extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200, response.text)
            self.assertEqual('yo yo', response.json)



#
# Copyright (C) 2016 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -*- coding: utf-8 -*-


from pecan import conf
from oslo_policy import policy


RULE_ADMIN_OR_SERVICE = 'rule:admin_or_service'
RULE_ADMIN_API = 'rule:admin_api'
RULE_ANY = 'rule:any'
RULE_NONE = 'rule:none'
RULE_OWNER_API = 'rule:owner_api'
RULE_ADMIN_OR_OWNER = 'rule:admin_or_owner'
RULE_PTL_API = 'rule:ptl_api'
RULE_CORE_API = 'rule:core_api'
RULE_DEV_API = 'rule:dev_api'
RULE_CONTRIBUTOR_API = 'rule:contributor_api'
RULE_AUTHENTICATED_API = 'rule:authenticated_api'

# TODO this value should be in the conf
# TODO keycloak stores usernames in lowercase ...
SERVICE_USER = 'SF_SERVICE_USER'

try:
    admin_account = conf.admin['name']
except AttributeError:
    admin_account = 'admin'


@policy.register('is_authenticated')
class IsAuthenticatedCheck(policy.Check):
    """An explicit check for user authentication."""

    def __call__(self, target, creds, enforcer):
        """Determine whether username is set or not."""

        return bool(creds.get('username'))


@policy.register('group')
class GroupCheck(policy.Check):
    """Check that there is a matching group in the ``creds`` dict."""

    def __call__(self, target, creds, enforcer):
        try:
            match = self.match % target
        except KeyError:
            # While doing RoleCheck if key not
            # present in Target return false
            return False
        if 'groups' in creds:
            return match.lower() in [x.lower() for x in creds['groups']]
        return False


rules = [
    policy.RuleDefault('is_admin', 'username:%s' % admin_account),
    policy.RuleDefault('is_service_upper',
                       'username:%s' % SERVICE_USER),
    policy.RuleDefault('is_service_lower',
                       'username:%s' % SERVICE_USER.lower()),
    policy.RuleDefault('is_service',
                       'rule:is_service_lower or rule:is_service_upper'),
    policy.RuleDefault('admin_or_service',
                       'rule:is_admin or rule:is_service'),
    policy.RuleDefault('admin_api', 'rule:is_admin'),
    policy.RuleDefault('is_owner', 'username:%(username)s'),
    policy.RuleDefault('owner_api', 'rule:is_owner'),
    policy.RuleDefault('admin_or_owner',
                       'rule:is_admin or rule:is_owner'),
    policy.RuleDefault('is_ptl', 'group:%(project)s-ptl'),
    policy.RuleDefault('is_core', 'group:%(project)s-core'),
    policy.RuleDefault('is_dev', 'group:%(project)s-dev'),
    policy.RuleDefault('ptl_api', 'rule:is_ptl'),
    policy.RuleDefault('core_api', 'rule:is_core'),
    policy.RuleDefault('dev_api', 'rule:is_dev'),
    policy.RuleDefault('contributor_api',
                       'rule:ptl_api or rule:core_api or rule:dev_api'),
    policy.RuleDefault('authenticated_api',
                       'is_authenticated:True'),
    policy.RuleDefault('any', '@'),
    policy.RuleDefault('none', '!'),
]


def list_rules():
    return rules



#
# Copyright (C) 2020 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from urllib.parse import urlparse
import requests

from managesf.lib.common import get_cauth_info
from managesf.lib.auth.base import BaseAuthenticator


class CAuthAuthenticator(BaseAuthenticator):

    def get_request_args(self, remote_gateway,
                         username=None, password=None,
                         github_access_token=None,
                         api_key=None,
                         use_ssl=True,
                         verify=True, **kwargs):
        req_args = {}
        if use_ssl is False:
            use_ssl = True
        if urlparse(remote_gateway).scheme == '':
            remote_gateway = "https://%s" % remote_gateway
        cauth_info = get_cauth_info(remote_gateway, verify)
        url = "%s/auth/login" % remote_gateway
        auth_params = {'back': '/',
                       'args': {}, }
        methods = cauth_info['service']['auth_methods']
        if (username and password and ('Password' in methods)):
            auth_params['args'] = {'username': username,
                                   'password': password}
            auth_params['method'] = 'Password'
        elif (github_access_token and
              ('GithubPersonalAccessToken' in methods)):
            auth_params['args'] = {'token': github_access_token}
            auth_params['method'] = 'GithubPersonalAccessToken'
        elif (api_key and ('APIKey' in methods)):
            auth_params['args'] = {'api_key': api_key}
            auth_params['method'] = 'APIKey'
        else:
            m = "Missing credentials (accepted auth methods: %s)"
            methods = ','.join(methods)
            raise ValueError(m % methods)
        header = {'Content-Type': 'application/json'}
        resp = requests.post(url, json.dumps(auth_params, sort_keys=True),
                             allow_redirects=False,
                             verify=verify,
                             headers=header)
        cookie = resp.cookies.get('auth_pubtkt', '')
        req_args['cookies'] = {'auth_pubtkt': cookie}
        req_args['headers'] = {'X-Remote-User': username}
        return req_args



#
# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from pecan import make_app
from pecan.deploy import deploy
from wsgiref import simple_server
from managesf import model


def setup_app(config):

    model.init_model()
    app_conf = dict(config.app)

    return make_app(
        app_conf.pop('root'),
        logging=getattr(config, 'logging', {}),
        **app_conf
    )


def main():
    app = deploy("/etc/managesf/config.py")
    host = app.config['server']['host']
    port = app.config['server']['port']
    srv = simple_server.make_server(host, port, app)
    srv.serve_forever()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase

from managesf.tests import dummy_conf
from managesf.model.yamlbkd.resources.project import ProjectOps


class ProjectOpsTest(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()

    def test_create(self):
        p = ProjectOps(self.conf, None)
        logs = p.create()
        self.assertEqual(len(logs), 0)

    def test_update(self):
        p = ProjectOps(self.conf, None)
        logs = p.update()
        self.assertEqual(len(logs), 0)

    def test_delete(self):
        p = ProjectOps(self.conf, None)
        logs = p.delete()
        self.assertEqual(len(logs), 0)

    def test_extra_validations(self):
        new = {
            'resources': {
                'projects': {
                    'p1': {
                        'source-repositories': []
                    }
                }
            }
        }
        p = ProjectOps(self.conf, new)
        kwargs = {'name': 'p1'}
        logs = p.extra_validations(**kwargs)
        self.assertEqual(len(logs), 0)



#
# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from pecan import conf  # noqa
from sqlalchemy import create_engine, Column, String, Unicode, UnicodeText
from sqlalchemy import Boolean, Integer, exc, event
from sqlalchemy import ForeignKey, UniqueConstraint
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.ext.declarative import declarative_base
from contextlib import contextmanager


Base = declarative_base()
engine = None


logger = logging.getLogger(__name__)


def row2dict(row):
    ret = {}
    for column in row.__table__.columns:
        ret[column.name] = getattr(row, column.name)
        # TODO: Fix test and remove bellow hack!
        if not isinstance(ret[column.name], (str, bytes)) and \
           not isinstance(ret[column.name], bool):
            ret[column.name] = str(ret[column.name])
    return ret


class User(Base):
    __tablename__ = 'users'
    username = Column(Unicode(255), primary_key=True)
    fullname = Column(Unicode(255), nullable=False)
    email = Column(String(255), nullable=False)
    hashed_password = Column(String(255), nullable=False)
    sshkey = Column(String(1023), nullable=True)


def checkout_listener(dbapi_con, con_record, con_proxy):
    try:
        try:
            dbapi_con.ping(False)
        except TypeError:
            dbapi_con.ping()
    except dbapi_con.OperationalError as e:
        if e.args[0] in (2006,   # MySQL server has gone away
                         2013,   # Lost connection to server during query
                         2055):  # Lost connection to server
            # caught by pool, which will retry with a new connection
            raise exc.DisconnectionError()
        else:
            raise


class SFUser(Base):
    __tablename__ = 'SF_USERS'
    id = Column(Integer(), primary_key=True)
    username = Column(Unicode(255), nullable=False, unique=True)
    fullname = Column(Unicode(255), nullable=True)
    # Gerrit requires email unicity
    email = Column(String(255), nullable=False, unique=True)
    cauth_id = Column(Integer(), nullable=False)
    idp_sync = Column(Boolean(), default=True)


class SFUserServiceMapping(Base):
    __tablename__ = 'SF_USERS_SERVICES_MAPPING'
    # needed for constraint definition, not actually used
    id = Column(Integer(), primary_key=True)
    sf_user_id = Column(Integer(), ForeignKey('SF_USERS.id'),
                        nullable=False)
    # will simply store the plugin name
    service = Column(String(255), nullable=False)
    # for extended future compatibility, don't limit to integers
    service_user_id = Column(String(255), nullable=False)
    __table_args__ = (UniqueConstraint('sf_user_id',
                                       'service',
                                       'service_user_id',
                                       name='unique_service_user'), )


class SFUserCRUD:
    def set_service_mapping(self, sf_user_id, service, service_user_id):
        with session_scope() as session:
            r = SFUserServiceMapping(sf_user_id=sf_user_id,
                                     service=service,
                                     service_user_id=service_user_id)
            session.add(r)

    def get_service_mapping(self, service, sf_user_id):
        with session_scope() as session:
            filtering = {'service': service,
                         'sf_user_id': sf_user_id}
            try:
                r = session.query(SFUserServiceMapping).filter_by(**filtering)
                return r.one().service_user_id
            except MultipleResultsFound:
                msg = 'Too many mappings for user #%s on service %s'
                raise KeyError(msg % (sf_user_id, service))
            except NoResultFound:
                return None

    def get_user_mapping(self, service, service_user_id):
        with session_scope() as session:
            filtering = {'service': service,
                         'service_user_id': service_user_id}
            try:
                r = session.query(SFUserServiceMapping).filter_by(**filtering)
                return r.one().service_user_id
            except MultipleResultsFound:
                msg = 'Too many mappings for service %s\'s user #%s'
                raise KeyError(msg % (service_user_id, service))
            except NoResultFound:
                return None

    def delete_service_mapping(self, sf_user_id,
                               service=None, service_user_id=None):
        with session_scope() as session:
            filtering = {'sf_user_id': sf_user_id}
            if service:
                filtering['service'] = service
            if service_user_id:
                filtering['service_user_id'] = service_user_id
            m = session.query(SFUserServiceMapping).filter_by(**filtering)
            m.delete(synchronize_session=False)

    def get(self, id=None, username=None, email=None,
            fullname=None, cauth_id=None):
        with session_scope() as session:
            if (id or username or email or fullname or cauth_id):
                filtering = {}
                if id:
                    filtering['id'] = id
                if username:
                    filtering['username'] = username
                if email:
                    filtering['email'] = email
                if fullname:
                    filtering['fullname'] = fullname
                if cauth_id:
                    filtering['cauth_id'] = cauth_id
                try:
                    ret = session.query(SFUser).filter_by(**filtering).one()
                    return row2dict(ret)
                except MultipleResultsFound:
                    # TODO(mhu) find a better Error
                    raise KeyError('search returned more than one result')
                except NoResultFound:
                    return {}
            else:
                # all()
                all = [row2dict(ret) for ret in session.query(SFUser)]
                return all

    def update(self, id, username=None, email=None,
               fullname=None, cauth_id=None, idp_sync=None):
        with session_scope() as session:
            try:
                ret = session.query(SFUser).filter_by(id=id).one()
                if username:
                    ret.username = username
                if email:
                    ret.email = email
                if fullname:
                    ret.fullname = fullname
                if cauth_id:
                    ret.cauth_id = cauth_id
                if idp_sync is not None:
                    ret.idp_sync = idp_sync
                session.commit()
            except MultipleResultsFound:
                msg = 'SF_USERS table has multiple row with the same id!'
                logger.error(msg)
                raise KeyError(msg)
            except NoResultFound:
                logger.warn("Could not update user %s: not found" % id)
                return

    def create(self, username, email,
               fullname, cauth_id=None):
        with session_scope() as session:
            if username and email and fullname:
                # assign a dummy value in case we lack the information
                # as is likely to happen when migrating from a previous version
                # TODO(mhu) remove these for version n+2
                cid = cauth_id or -1
                user = SFUser(username=username,
                              email=email,
                              fullname=fullname,
                              cauth_id=cid)
                session.add(user)
                session.commit()
                return user.id
            else:
                msg = "Missing info required for user creation: %s|%s|%s"
                raise KeyError(msg % (username, email, fullname))

    def delete(self, id=None, username=None, email=None,
               fullname=None, cauth_id=None):
        with session_scope() as session:
            filtering = {}
            if id:
                filtering['id'] = id
            if username:
                filtering['username'] = username
            if email:
                filtering['email'] = email
            if fullname:
                filtering['fullname'] = fullname
            if cauth_id:
                filtering['cauth_id'] = cauth_id
            try:
                ret = session.query(SFUser).filter_by(**filtering).one()
                session.delete(ret)
                session.commit()
                return True
            except MultipleResultsFound:
                # TODO(mhu) find a better Error
                raise KeyError('Too many candidates for deletion')
            except NoResultFound:
                return False


class NodepoolImageUpdate(Base):
    __tablename__ = 'NODEPOOL_IMAGE_UPDATES'
    id = Column(Integer(), primary_key=True)
    status = Column(String(255), default="IN_PROGRESS")
    provider = Column(String(1024), nullable=False)
    image = Column(String(1024), nullable=False)
    exit_code = Column(Integer(), default=-1)
    stderr = Column(UnicodeText(), default="")
    output = Column(UnicodeText(4294967295), default="")


class ImageUpdatesCRUD():
    def create(self, provider, image):
        with session_scope() as session:
            if provider and image:
                img_update = NodepoolImageUpdate(
                              provider=provider,
                              image=image)
                session.add(img_update)
                session.commit()
                return img_update.id
            else:
                msg = "Missing info required for image update: %s|%s"
                raise KeyError(msg % (provider, image))

    def update(self, id, status=None, exit_code=None,
               output=None, stderr=None):
        with session_scope() as session:
            try:
                u = session.query(NodepoolImageUpdate).filter_by(id=id).one()
                if status:
                    u.status = status
                if exit_code:
                    u.exit_code = int(exit_code)
                if output:
                    u.output = output
                if stderr:
                    u.stderr = stderr
                session.commit()
            except NoResultFound:
                logger.warn("Could not update image-update %s: not found" % id)
                return

    def get(self, id):
        with session_scope() as session:
            # TODO(mhu) Lookup by images, providers, statuses if needed?
            try:
                u = session.query(NodepoolImageUpdate).filter_by(id=id).one()
                return row2dict(u)
            except NoResultFound:
                return {}


def init_model():
    c = dict(conf.sqlalchemy)
    url = c.pop('url')
    if url.startswith('mysql://'):
        url = url.replace('mysql://', 'mysql+pymysql://')
    if url.startswith('mysql') and not url.endswith('?charset=utf8'):
        url += '?charset=utf8'
    globals()['engine'] = create_engine(url, pool_recycle=600, **c)
    if url.startswith('mysql'):
        event.listen(engine, 'checkout', checkout_listener)
    Base.metadata.create_all(engine)


def start_session():
    Base.metadata.bind = engine
    dbsession = sessionmaker(bind=engine)
    session = dbsession()
    return session


@contextmanager
def session_scope():
    session = start_session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def add_user(user):
    """ Add a user in the database
        return Boolean
    """
    try:
        with session_scope() as session:
            u = User(**user)
            session.add(u)
            return True, None
    except exc.IntegrityError as e:
        return False, str(e)


def get_user(username):
    """ Fetch a user by its username
        return user dict or False if not found
    """
    try:
        with session_scope() as session:
            ret = session.query(User).filter_by(username=username).one()
            return row2dict(ret)
    except NoResultFound:
        return False


def delete_user(username):
    """ Delete a user by its username
        return True if deleted or False if not found
    """
    with session_scope() as session:
        ret = session.query(User).filter_by(username=username).delete()
        return bool(ret)


def update_user(username, infos):
    """ Update a user by its username
        arg infos: Dict
        return True if deleted or False if not found
    """
    with session_scope() as session:
        user = session.query(User)
        ret = user.filter_by(username=username).update(infos)
        return bool(ret)



#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging
import os.path
import re

from pecan import conf
from pecan import request, response, abort, expose
from pecan.rest import RestController

from managesf.model.yamlbkd.engine import SFResourceBackendEngine
from managesf import policy
# TODO do it with v2
from managesf.model import SFUserCRUD
from git.exc import GitCommandError


logger = logging.getLogger(__name__)


# TODO move to managesf.api once users API is started
def get_user_groups(username):
    user_email = SFUserCRUD().get(username=username).get('email')
    logger.info('Found email %s for username %s' % (user_email, username))
    resources_engine = SFResourceBackendEngine(
        os.path.join(conf.resources['workdir'], 'read'),
        conf.resources['subdir'])
    try:
        resources = resources_engine.get(
            conf.resources['master_repo'], 'master')
    except GitCommandError:
        logger.info("Unable to read groups from the resources engine.")
        logger.info("It is probably because we are boostrapping SF.")
        return []
    groups = resources['resources'].get('groups', {})
    return [g for g in groups if user_email in groups[g]['members']]


def authorize(rule_name, target):
    if not request.remote_user:
        request.remote_user = request.headers.get('X-Remote-User')
    credentials = {'username': request.remote_user, 'groups': []}
    if request.remote_user:
        credentials['groups'] = get_user_groups(request.remote_user)
    return policy.authorize(rule_name, target, credentials)


class APIv2RestController(RestController):
    def __init__(self, *args, **kwargs):
        super(APIv2RestController, self).__init__(*args, **kwargs)
        self._logger = logging.getLogger(
            'managesf.v2.controllers.%s' % self.__class__.__name__)


class APIv2RestProxyController(APIv2RestController):
    manager = None
    policies_map = {'get .+/path/to/(?P<x>.+)/command': 'managesf.policy.name'}

    def _find_policy(self, lookup):
        """Find policy according to REST path."""
        for expr in self.policies_map:
            regex = re.compile(expr)
            if regex.search(lookup):
                target_elements = regex.search(lookup).groupdict()
                return {'policy': self.policies_map[expr],
                        'target_elements': target_elements}
        return {}

    def _policy_target(self, verb, target_elements, *args, **kwargs):
        # override me
        target = {}
        return target

    # This makes the assumption that backend services always return JSON.
    # This is true for all of them except gerrit, which will not be covered
    # this way.
    def _do(self, verb):

        def action(*args, **kwargs):
            if not self.manager:
                return abort(404,
                             detail='This service is not configured.')
            path = request.path
            lookup = ("%s %s" % (verb, path))
            pol_scan = self._find_policy(lookup)
            pol, target_elements = None, {}
            if pol_scan:
                pol = pol_scan['policy']
                target_elements = pol_scan['target_elements']
            if not kwargs and request.content_length:
                if 'json' in request.content_type:
                    kwargs = request.json
                else:
                    kwargs = request.params
            target = self._policy_target(verb, target_elements,
                                         *args, **kwargs)
            if not pol:
                # Unknown endpoint, default behavior is to forbid access
                pol = 'rule:none'
            if not authorize(pol, target=target):
                return abort(401,
                             detail='Failure to comply with policy %s' % pol)
            # HACK The RestController's routing method seems to discard
            # extensions on the last remainder. This is a dirty fix
            full_path = request.path_url
            last_arg = full_path.split('/')[-1]
            if last_arg and args[-1] != last_arg:
                args = args[:-1] + (last_arg, )
            if request.content_length and 'json' in request.content_type:
                proxied_response = getattr(self.manager, verb)(
                    *args, json=kwargs)
                self._logger.debug(
                    "calling passthrough manager with "
                    "args: '%s', json: '%s'" % (args, kwargs))
            elif kwargs:
                proxied_response = getattr(self.manager, verb)(
                    *args, params=kwargs)
                self._logger.debug(
                    "calling passthrough manager with "
                    "args: '%s', params: '%s'" % (args, kwargs))
            else:
                proxied_response = getattr(self.manager, verb)(*args)
            response.status = proxied_response.status_code
            if int(proxied_response.status_code) > 399:
                response.text = proxied_response.text
                return abort(proxied_response.status_code)
            else:
                return proxied_response.json()

        return action

    @expose('json')
    def get(self, *args, **kwargs):
        return self._do('get')(*args, **kwargs)

    @expose('json')
    def post(self, *args, **kwargs):
        return self._do('post')(*args, **kwargs)

    @expose('json')
    def put(self, *args, **kwargs):
        return self._do('put')(*args, **kwargs)

    @expose('json')
    def delete(self, *args, **kwargs):
        return self._do('delete')(*args, **kwargs)



#
# Copyright (C) 2020 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import requests # noqa
from urllib.parse import urlparse

from managesf.lib.auth.base import BaseAuthenticator


class KeycloakAuthenticator(BaseAuthenticator):

    def get_request_args(self, remote_gateway,
                         username=None, password=None,
                         verify=True,
                         **kwargs):
        # fetch the well-known conf and endpoint
        # TODO we should have a manageSF endpoint that gives us the endpoint
        # rather than guessing it this way.
        if urlparse(remote_gateway).scheme == '':
            remote_gateway = "https://%s" % remote_gateway
        # TODO realm should be a parameter?
        conf_url = "%s/auth/realms/sf/.well-known/openid-configuration"
        kc_conf = requests.get(conf_url % remote_gateway).json()
        token_endpoint = kc_conf.get("token_endpoint")
        if token_endpoint is None:
            raise Exception("No token endpoint found")
        # get the token
        data = {
            'username': username,
            'password': password,
            'grant_type': 'password',
            'client_id': 'managesf',
        }
        token_request = requests.post(
            token_endpoint,
            data,
            verify=verify)
        try:
            token = token_request.json()['access_token']
        except Exception:
            raise Exception(
                'Failure when fetching token: %s' % token_request.status_code)
        return {'headers': {'Authorization': 'bearer %s' % token}}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import yaml
try:
    from yaml import CSafeLoader as YLoader
except ImportError:
    from yaml import SafeLoader as YLoader
import shutil

from unittest import TestCase
from mock import patch

from managesf.model.yamlbkd import yamlbackend
from managesf.tests import resources_test_utils as rtu


class YAMLBackendTest(TestCase):
    def setUp(self):
        self.db_path = []

    def tearDown(self):
        for db_path in self.db_path:
            if os.path.isdir(db_path):
                shutil.rmtree(db_path)
            else:
                os.unlink(db_path)

    def test_load_valid_db_data(self):
        # Prepare a GIT repo with content
        repo_path = rtu.prepare_git_repo(self.db_path)
        # Add a file of data
        data = {'resources': {'repos': {'repo1': {},
                                        },
                              'projects': {'project1': {},
                                           },
                              'groups': {'group1': {}
                                         },
                              'acls': {'acl1': {}
                                       },
                              }
                }
        rtu.add_yaml_data(repo_path, data)
        # Init the YAML DB
        clone_path, cache_path = rtu.prepare_db_env(self.db_path)
        db = yamlbackend.YAMLBackend("file://%s" % repo_path,
                                     "master", "resources",
                                     clone_path, cache_path)
        self.assertIn('project1', db.get_data()['resources']['projects'])
        # Add another file of data
        data = {'resources': {'repos': {'repo2': {},
                                        },
                              'projects': {'project2': {},
                                           },
                              'groups': {'group2': {}
                                         },
                              'acls': {'acl2': {}
                                       },
                              }
                }
        rtu.add_yaml_data(repo_path, data)
        db.refresh()
        project_ids = db.get_data()['resources']['projects'].keys()
        self.assertIn('project1', project_ids)
        self.assertIn('project2', project_ids)
        self.assertEqual(len(project_ids), 2)
        # Add another file of data for another resource
        data = {'resources': {'groups': {
                'id': {'name': 'resource_a'}
                }}}
        rtu.add_yaml_data(repo_path, data)
        db.refresh()
        group_ids = db.get_data()['resources']['groups'].keys()
        self.assertIn('id', group_ids)

    def test_load_invalid_db_data(self):
        # Prepare a GIT repo with content
        repo_path = rtu.prepare_git_repo(self.db_path)
        # Add a file of invalid data
        data = {'resources': {'projects': {
                'id1': {'name': 'resource_a'},
                }}}
        data2 = {'resources': {'projects': {
                 'id1': {'name': 'resource_b'},
                 }}}
        rtu.add_yaml_data(repo_path, data)
        rtu.add_yaml_data(repo_path, data2)
        # Init the YAML DB
        clone_path, cache_path = rtu.prepare_db_env(self.db_path)
        with self.assertRaises(yamlbackend.YAMLDBException):
            yamlbackend.YAMLBackend("file://%s" % repo_path,
                                    "master", "resources",
                                    clone_path,
                                    cache_path)

    def test_db_data_struct(self):
        # Init the DB with valid data
        repo_path = rtu.prepare_git_repo(self.db_path)
        data = {'resources': {'repos': {'repo1': {},
                                        },
                              'projects': {'project1': {},
                                           },
                              'groups': {'group1': {}
                                         },
                              'acls': {'acl1': {}
                                       },
                              }
                }
        rtu.add_yaml_data(repo_path, data)
        clone_path, cache_path = rtu.prepare_db_env(self.db_path)
        db = yamlbackend.YAMLBackend("file://%s" % repo_path,
                                     "master", "resources",
                                     clone_path,
                                     cache_path)
        # Try to validate a bunch a invalid data
        rids = {}
        for data in [
            42,
            [],
            {'wrong': {}},
            {'resources': {4: []}},
            {'resources': {'projects': [None]}},
            {'resources': {'projects': {None: []}}},
            {'resources': {'projects': {'id': []}}},
            {'resources': {'projects': {'id': []}, 'groups': {'id': []}}},
        ]:
            self.assertRaises(yamlbackend.YAMLDBException,
                              db.validate, data, rids)

    def test_db_data_struct_extra(self):
        # Init the DB with valid data
        repo_path = rtu.prepare_git_repo(self.db_path)

        data = """
'
   resources:
projects: {}
"""
        rtu.add_yaml_data(repo_path, data, free_style=True)
        clone_path, cache_path = rtu.prepare_db_env(self.db_path)
        self.assertRaises(yamlbackend.YAMLDBException,
                          yamlbackend.YAMLBackend,
                          "file://%s" % repo_path,
                          "master", "resources",
                          clone_path, cache_path)

    def test_db_cache(self):
        # Init the DB with validate data
        repo_path = rtu.prepare_git_repo(self.db_path)
        data = {'resources': {'repos': {'repo1': {},
                                        },
                              'projects': {'project1': {},
                                           },
                              'groups': {'group1': {}
                                         },
                              'acls': {'acl1': {}
                                       },
                              }
                }
        rtu.add_yaml_data(repo_path, data)
        clone_path, cache_path = rtu.prepare_db_env(self.db_path)
        db = yamlbackend.YAMLBackend("file://%s" % repo_path,
                                     "master", "resources",
                                     clone_path,
                                     cache_path)
        # Some verification about the cache content
        repo_hash = db._get_repo_hash()
        cache_hash = db._get_cache_hash()
        self.assertEqual(repo_hash, cache_hash)
        cached_data = yaml.load(open(db.cache_path),
                                Loader=YLoader)
        self.assertIn('projects', cached_data['resources'])
        # Add more data in the db
        data = {'resources': {'groups': {'group2': {}}}}
        rtu.add_yaml_data(repo_path, data)
        db = yamlbackend.YAMLBackend("file://%s" % repo_path,
                                     "master", "resources",
                                     clone_path,
                                     cache_path)
        repo_hash2 = db._get_repo_hash()
        cache_hash2 = db._get_cache_hash()
        self.assertEqual(repo_hash2, cache_hash2)
        self.assertNotEqual(cache_hash, cache_hash2)
        cached_data2 = yaml.load(open(db.cache_path),
                                 Loader=YLoader)
        self.assertIn('projects', cached_data2['resources'])
        self.assertIn('groups', cached_data2['resources'])
        # Re-create the YAMLBackend instance whithout changed
        # in the upstream GIT repo
        with patch.object(yamlbackend.YAMLBackend, '_load_db') as ldb:
            with patch.object(yamlbackend.YAMLBackend, '_update_cache') as uc:
                db = yamlbackend.YAMLBackend("file://%s" % repo_path,
                                             "master", "resources",
                                             clone_path,
                                             cache_path)
        cache_hash3 = db._get_cache_hash()
        self.assertEqual(cache_hash3, cache_hash2)
        self.assertFalse(ldb.called or uc.called)


class MemoryYAMLBackendTest(TestCase):
    def setUp(self):
        self.db_path = []

    def tearDown(self):
        for db_path in self.db_path:
            if os.path.isdir(db_path):
                shutil.rmtree(db_path)
            else:
                os.unlink(db_path)

    def test_load_valid_db_data(self):
        f1 = {
            'resources': {
                'repos': {
                    'repo1': {},
                },
                'projects': {
                    'project1': {},
                },
                'groups': {
                    'group1': {}
                },
                'acls': {
                    'acl1': {}
                },
            }
        }
        f2 = {
            'resources': {
                'repos': {
                    'repo2': {},
                },
                'projects': {
                    'project2': {},
                },
                'groups': {
                    'group2': {}
                },
                'acls': {
                    'acl2': {}
                },
            }
        }
        f3 = {
            'resources': {
                'groups': {
                    'id': {
                        'name': 'resource_a'}
                }
            }
        }
        data = {
            'f1': f1,
            'f2': f2,
            'f3': f3}
        db = yamlbackend.MemoryYAMLBackend(data)
        project_ids = db.get_data()['resources']['projects'].keys()
        self.assertIn('project1', project_ids)
        self.assertIn('project2', project_ids)
        self.assertEqual(len(project_ids), 2)
        group_ids = db.get_data()['resources']['groups'].keys()
        self.assertIn('id', group_ids)

    def test_load_invalid_db_data(self):
        f1 = {
            'resources': {
                'projects': {
                    'id1': {
                        'name': 'resource_a'
                    },
                }
            }
        }
        f2 = {
            'resources': {
                'projects': {
                    'id1': {
                        'name': 'resource_b'
                    },
                 }
            }
        }
        data = {
            'f1': f1,
            'f2': f2,
        }
        with self.assertRaises(yamlbackend.YAMLDBException):
            yamlbackend.MemoryYAMLBackend(data)

    def test_db_data_struct(self):
        data = {
            'resources': {
                'repos': {
                    'repo1': {},
                },
                'projects': {
                    'project1': {},
                },
                'groups': {
                    'group1': {}
                },
                'acls': {
                    'acl1': {}
                },
            }
        }
        db = yamlbackend.MemoryYAMLBackend({'data': data})
        # Try to validate a bunch a invalid data
        rids = {}
        for data in [
            42,
            [],
            {'wrong': {}},
            {'resources': {4: []}},
            {'resources': {'projects': [None]}},
            {'resources': {'projects': {None: []}}},
            {'resources': {'projects': {'id': []}}},
            {'resources': {'projects': {'id': []}, 'groups': {'id': []}}},
        ]:
            self.assertRaises(yamlbackend.YAMLDBException,
                              db.validate, data, rids)

    def test_db_data_struct_extra(self):
        data = """
'
   resources:
projects: {}
"""
        self.assertRaises(yamlbackend.YAMLDBException,
                          yamlbackend.MemoryYAMLBackend,
                          {'data': data})



#!/usr/bin/env python
#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# import logging
import os.path

from managesf.api.v2 import resources
from managesf.model.yamlbkd.engine import SFResourceBackendEngine


"""Resources service plugin using manageSF's built-in resource engine."""


# logger = logging.getLogger(__name__)

class ResourcesManager(resources.ResourcesManager):
    def __init__(self, manager):
        super(ResourcesManager, self).__init__()
        self.manager = manager

    def get(self, **kwargs):
        engine = self.manager.get_engine('read')
        if kwargs.get('get_missing_resources'):
            return engine.get_missing_resources(self.manager.master_repo,
                                                'master')
        else:
            return engine.get(self.manager.master_repo, 'master')

    def create(self, **kwargs):
        if kwargs.get('data') is None:
            raise ValueError('Invalid request: missing "data"')
        data = kwargs['data']
        prev_ref = kwargs.get('prev_ref', 'master')
        engine = self.manager.get_engine('validate')
        status, logs = engine.validate_from_structured_data(
            self.manager.master_repo, prev_ref, data)
        return status, logs


class SFResourcesManager(resources.ResourcesServiceManager):

    _config_section = "resources"
    service_name = "manageSF"

    def __init__(self, conf):
        super(SFResourcesManager, self).__init__(conf)
        self.subdir = self.conf['subdir']
        self.workdir = self.conf['workdir']
        self.master_repo = self.conf['master_repo']
        self.resources = ResourcesManager(self)

    def get_engine(self, operation):
        """Returns a resource engine for the right operation.
        Valid operations: read, validate, apply"""
        if operation not in ['read', 'validate', 'apply']:
            raise ValueError('Unknown operation "%s"' % operation)
        engine = SFResourceBackendEngine(
            os.path.join(self.workdir, operation),
            self.subdir)
        return engine



#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import tempfile
import logging

from git.config import GitConfigParser

from managesf.model.yamlbkd.resource import BaseResource

logger = logging.getLogger(__name__)

KEYS_EXP_GROUP_REGEXS = (
    'owner',
    'read',
    'submit',
    'rebase',
    'forgeAuthor',
    'create',
    'push',
    'label-[^ ]+',
    'addPatchSet'
)

KEYS_EXP_BOOLEAN_VALUES = (
    'requireChangeId',
    'mergeContent',
    'requireContributorAgreement',
    'requireSignedOffBy',
)

SUBMIT_ACTION_VALUES = (
    'fast forward only',
    'merge if necessary',
    'rebase if necessary',
    'always merge',
    'cherry pick',
)


class ACLOps(object):

    def __init__(self, conf, new):
        self.conf = conf
        self.new = new

    def extra_validations(self, **kwargs):
        default_groups = ('Service Users',
                          'Non-Interactive Users',
                          'Administrators',
                          'Registered Users',
                          'Anonymous Users')
        logs = []
        acls = kwargs['file']
        groups = kwargs['groups']
        name = kwargs['name']

        fd, path = tempfile.mkstemp()
        os.close(fd)
        open(path, 'w').write(acls)
        # Validate the file has the right format
        try:
            c = GitConfigParser(path)
            c.read()
        except Exception as e:
            logger.exception("GitConfigParser failed %s" % e)
            logs.append(str(e))
        finally:
            os.unlink(path)

        # Verify the groups mentioned in the ACLs file are known
        group_names = set()
        for group_id in groups:
            name = self.new['resources']['groups'][group_id]['name']
            group_names.add(name)

        # Verify the validity of the ACL
        sections = [s for s in c.sections() if c != 'project']
        for section_name in sections:
            for k, v in c.items(section_name):
                if k in KEYS_EXP_BOOLEAN_VALUES:
                    if v.lower() not in ('true', 'false'):
                        logs.append(
                            "ACLs file section (%s), key (%s) expects "
                            "a valid boolean value (not: %s)" % (
                                section_name, k, v))
                    continue
                if any([re.match(rex, k) for rex in KEYS_EXP_GROUP_REGEXS]):
                    if k.startswith('label-'):
                        if not re.match('.+ group .+$', v):
                            logs.append(
                                "ACLs file section (%s), key (%s) expects "
                                "a note rule and a group to be specified "
                                "(not: %s)" % (section_name, k, v))
                    else:
                        if not re.match('(deny group|group) .+$', v):
                            logs.append(
                                "ACLs file section (%s), key (%s) expects "
                                "a group to be specified (not: %s)" % (
                                    section_name, k, v))
                if k == 'action' and section_name == 'submit':
                    if v.lower() not in SUBMIT_ACTION_VALUES:
                        logs.append(
                            "ACLs file section (%s), key (%s) expects "
                            "a valid submit strategy (not: %s)" % (
                                section_name, k, v))
                    continue
                m = re.search('.*group (.*)$', v)
                if m:
                    group_name = m.groups()[0]
                    if (group_name not in group_names and
                            group_name not in default_groups):
                        logs.append(
                            "ACLs file section (%s), key (%s) relies on an "
                            "unknown group name: %s" % (
                                section_name, k, group_name))

        # Try to detect an ACL that make a repository private
        if ('read = deny group Registered Users' in acls and
                'read = deny group Anonymous Users' in acls):
            # That a private ACLs then make sure repos that use it
            # use the private flag at project level
            # Find repos that use this ACL
            private_repos = []
            for repo, rdata in self.new['resources']['repos'].items():
                if rdata['acl'] == name:
                    private_repos.append(repo)
            # Find repos marked as private in projects
            # and remove them from the private_repos accu
            if private_repos:
                projects = self.new['resources']['projects'].values()
                for project in projects:
                    for sr in project['source-repositories']:
                        if isinstance(sr, dict):
                            sr_name = list(sr.keys())[0]
                            sr_attrs = sr[sr_name]
                            if sr_attrs.get('private') is True:
                                private_repos.remove(sr_name)
            # If some remains in the accu then the 'private' attribute
            # is missing on them
            if private_repos:
                logs.append(
                    "%s repositories use a private Gerrit ACL but are not"
                    " defined as private in a project" % private_repos)

        return logs


class ACL(BaseResource):

    MODEL_TYPE = 'acl'
    DESCRIPTION = ("The acl resource is used to store a Gerrit ACL. "
                   "The acl can be shared between multiple git repositories. "
                   "Group mentionned inside the acl file key must be "
                   "referenced by their ID under the groups key. Do not "
                   "provide the description entry in the acl file to keep "
                   "them shareable between git repositories if needed.")
    MODEL = {
        'name': (
            str,
            '.*',
            False,
            '',
            False,
            "The ACL name",
        ),
        'file': (
            str,
            '.*',
            True,
            None,
            True,
            "The Gerrit ACL content",
        ),
        'groups': (
            list,
            '.+',
            False,
            [],
            True,
            "The list of groups this ACL depends on",
        ),
    }
    PRIORITY = 30
    PRIMARY_KEY = 'file'
    CALLBACKS = {
        'update': lambda conf, new, kwargs: [],
        'create': lambda conf, new, kwargs: [],
        'delete': lambda conf, new, kwargs: [],
        'extra_validations': lambda conf, new, kwargs:
            ACLOps(conf, new).extra_validations(**kwargs),
        'get_all': lambda conf, new: ([], {}),
    }

    def get_deps(self, keyname=False):
        if keyname:
            return 'groups'
        return {'groups': set(self.resource['groups'])}

    def should_be_updated(self):
        """ An ACL depends on one or more groups
        but a group update does not need to force
        the engine to call the ACL update callback.
        """
        return False



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import TestCase

from managesf.model.yamlbkd.resources.gitacls import ACLOps


class TestACLOps(TestCase):

    def test_gerrit_plugin_config(self):
        """Test that adding extra plugin config options in project.config
        does not break validation"""
        new = {
            'resources': {
                'groups': {
                    'mygid': {
                        'name': 'coders',
                        'namespace': '',
                        'members': ['body@sftests.com'],
                    }
                }
            }
        }
        kwargs = {'file': """[project]
\tdescription = "My awesome project"
[access "refs/*"]
\tread = group coders
\tlabel-Code-Review = -2..+2 group coders
[plugin "reviewers-by-blame"]
\tmaxReviewers = 2
\tignoreDrafts = true
\tignoreSubjectRegEx = WIP(.*)
""",
                  'groups': ['mygid'],
                  'name': 'myacl'}
        o = ACLOps(None, new)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 0)

    def test_extra_validations(self):
        kwargs = {'file': 'invalid ACLs !',
                  'groups': [],
                  'name': 'myacl'}
        o = ACLOps(None, None)
        logs = o.extra_validations(**kwargs)
        self.assertTrue(logs[0].startswith(
            "File contains no section headers."))
        self.assertEqual(len(logs), 1)

        kwargs = {'file': '',
                  'groups': [],
                  'name': 'myacl'}
        o = ACLOps(None, None)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 0)

        new = {
            'resources': {
                'groups': {
                    'mygid': {
                        'name': 'coders',
                        'namespace': '',
                        'members': ['body@sftests.com'],
                    }
                }
            }
        }

        kwargs = {'file': """[project]
\tdescription = "My awesome project"
[access "refs/*"]
\tread = coders
""",
                  'groups': ['mygid'],
                  'name': 'myacl'}

        o = ACLOps(None, new)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 1)
        self.assertIn('ACLs file section (access "refs/*"), key '
                      '(read) expects a group to be specified (not: coders)',
                      logs)

        kwargs = {'file': """[project]
\tdescription = "My awesome project"
[access "refs/*"]
\tlabel-Code-Review = -2..+2 coders
""",
                  'groups': ['mygid'],
                  'name': 'myacl'}

        o = ACLOps(None, new)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 1)
        self.assertIn('ACLs file section (access "refs/*"), key '
                      '(label-Code-Review) expects a note rule and '
                      'a group to be specified (not: -2..+2 coders)',
                      logs)

        kwargs = {'file': """[project]
\tdescription = "My awesome project"
[access "refs/*"]
\tread = group coders
\tlabel-Code-Review = -2..+2 group coders
""",
                  'groups': ['mygid'],
                  'name': 'myacl'}
        o = ACLOps(None, new)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 0)

        new = {
            'resources': {
                'groups': {
                    'mygid': {
                        'name': 'pchitt',
                        'namespace': '',
                        'members': ['body@sftests.com'],
                    }
                }
            }
        }
        o = ACLOps(None, new)
        logs = o.extra_validations(**kwargs)
        self.assertEqual(len(logs), 2)
        self.assertIn('ACLs file section (access "refs/*"), key '
                      '(read) relies on an unknown group name: coders',
                      logs)
        self.assertIn('ACLs file section (access "refs/*"), key '
                      '(label-Code-Review) relies on an unknown '
                      'group name: coders',
                      logs)



# Copyright (C) 2018 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import yaml
from unittest import TestCase

from managesf.controllers.api.v2.configurations import ZuulTenantsLoad
from managesf.controllers.api.v2.configurations import NodepoolConf
from managesf.controllers.api.v2.configurations import HoundConf
from managesf.controllers.api.v2.configurations import CauthConf


class ZuulTenantsLoadTests(TestCase):
    def assertItemsEqual(self, *args, **kwargs):
        try:
            super(ZuulTenantsLoadTests, self).assertItemsEqual(*args, **kwargs)
        except Exception:  # python 3
            self.assertCountEqual(*args, **kwargs)

    def test_merge_tenant_from_flat_files(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenants_data = """
        - tenant:
            name: local
            source:
              gerrit:
                config-projects:
                  - common-config
        - tenant:
            name: local
            max-nodes-per-job: 5
            source:
              gerrit:
                untrusted-projects:
                  - repo1
                  - repo2
        """
        final_tenants = {}
        projects_list = {}
        tenants = yaml.safe_load(tenants_data)
        for tenant in tenants:
            ztl.merge_tenant_from_data(
                final_tenants, tenant, '/data', 'local', projects_list)
        projects_list_expected = {'local': ['common-config', 'repo1', 'repo2']}
        self.assertItemsEqual(
            projects_list['local'], projects_list_expected['local'])
        final_tenants = ztl.final_tenant_merge(final_tenants)
        expected = {
            'tenant': {
                'admin-rules': [
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN',
                ],
                'name': 'local',
                'max-nodes-per-job': 5,
                'source': {
                    'gerrit': {
                        'config-projects': ['common-config'],
                        'untrusted-projects': ['repo1', 'repo2']
                        }
                    }
                }
            }
        self.assertDictEqual(final_tenants[0], expected)
        self.assertEqual(len(final_tenants), 1)

    def test_merge_tenant_from_flat_files_with_admin_rules(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenants_data = """
        - tenant:
            name: local
            admin-rules:
              - rule1
              - rule2
            source:
              gerrit:
                config-projects:
                  - common-config
        - tenant:
            name: local
            max-nodes-per-job: 5
            source:
              gerrit:
                untrusted-projects:
                  - repo1
                  - repo2
        - authorization-rule:
            name: rule1
            conditions:
              - iss: some_iss
        - authorization-rule:
            name: rule2
            conditions:
              - email: some@email.com
        """
        final_tenants = {}
        final_rules = {}
        projects_list = {}
        tenants = yaml.safe_load(tenants_data)
        for data in tenants:
            if 'tenant' in data.keys():
                ztl.merge_tenant_from_data(
                    final_tenants, data, '/data', 'local', projects_list)
            else:
                ztl.merge_auth_rule_from_data(
                    final_rules, data)
        projects_list_expected = {'local': ['common-config', 'repo1', 'repo2']}
        self.assertItemsEqual(
            projects_list['local'], projects_list_expected['local'])
        final_tenants = ztl.final_tenant_merge(final_tenants)
        final_rules = ztl.merge_auth_rules(final_rules)
        expected = {
            'tenant': {
                'name': 'local',
                'admin-rules': [
                    'rule1',
                    'rule2',
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN',
                ],
                'max-nodes-per-job': 5,
                'source': {
                    'gerrit': {
                        'config-projects': ['common-config'],
                        'untrusted-projects': ['repo1', 'repo2']
                        }
                    }
                }
            }
        expected_rules = [
            {'authorization-rule': {
                'name': 'rule1',
                'conditions': [
                    {'iss': 'some_iss'}
                ]
            }},
            {'authorization-rule': {
                'name': 'rule2',
                'conditions': [
                    {'email': 'some@email.com'}
                ]
            }},
            {'authorization-rule': {
                'name': '__SF_DEFAULT_ADMIN',
                'conditions': [
                    {'username': 'admin'},
                    {'roles': 'zuul_admin'},
                ]
            }},
            {'authorization-rule': {
                'name': '__SF_TENANT_ZUUL_ADMIN',
                'conditions': [
                    {'roles': '{tenant.name}_zuul_admin'},
                ]
            }},
        ]
        self.assertDictEqual(final_tenants[0], expected)
        self.assertTrue(
            all(r in final_rules for r in expected_rules)
        )

    def test_merge_tenant_from_flat_files_cannot_override_default_rule(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenants_data = """
        - tenant:
            name: local
            source:
              gerrit:
                config-projects:
                  - common-config
        - tenant:
            name: local
            max-nodes-per-job: 5
            source:
              gerrit:
                untrusted-projects:
                  - repo1
                  - repo2
        - authorization-rule:
            name: __SF_DEFAULT_ADMIN
            conditions:
              - iss: some_iss
        """
        final_tenants = {}
        final_rules = {}
        projects_list = {}
        tenants = yaml.safe_load(tenants_data)
        for data in tenants:
            if 'tenant' in data.keys():
                ztl.merge_tenant_from_data(
                    final_tenants, data, '/data', 'local', projects_list)
            else:
                ztl.merge_auth_rule_from_data(
                    final_rules, data)
        projects_list_expected = {'local': ['common-config', 'repo1', 'repo2']}
        self.assertItemsEqual(
            projects_list['local'], projects_list_expected['local'])
        final_tenants = ztl.final_tenant_merge(final_tenants)
        final_rules = ztl.merge_auth_rules(final_rules)
        expected = {
            'tenant': {
                'name': 'local',
                'admin-rules': [
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN',
                ],
                'max-nodes-per-job': 5,
                'source': {
                    'gerrit': {
                        'config-projects': ['common-config'],
                        'untrusted-projects': ['repo1', 'repo2']
                        }
                    }
                }
            }
        expected_rules = [
            {'authorization-rule': {
                'name': '__SF_DEFAULT_ADMIN',
                'conditions': [
                    {'username': 'admin'},
                    {'roles': 'zuul_admin'},
                ]
            }},
            {'authorization-rule': {
                'name': '__SF_TENANT_ZUUL_ADMIN',
                'conditions': [
                    {'roles': '{tenant.name}_zuul_admin'},
                ]
            }},
        ]
        self.assertDictEqual(final_tenants[0], expected)
        self.assertTrue(expected_rules[0] in final_rules)

    def test_merge_tenant_error_from_flat_files(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenant_data = """
        - tenant:
            name: local
            source:
              gerrit:
                config-projects:
                  - config
        - tenant:
            name: local
            source:
              gerrit:
                config-projects:
                  - config
        """
        final_tenants = {}
        projects_list = {}
        tenants = yaml.safe_load(tenant_data)
        ztl.merge_tenant_from_data(
            final_tenants, tenants[0], '/data', 'local', projects_list)
        projects_list_expected = {'local': ['config']}
        self.assertItemsEqual(
            projects_list['local'], projects_list_expected['local'])
        with self.assertRaises(RuntimeError) as ctx:
            ztl.merge_tenant_from_data(
                final_tenants, tenants[1], '/data', 'local',  projects_list)
        self.assertEqual(
            str(ctx.exception),
            '/data: define existing project config for tenant local')

    def test_merge_tenants_from_flat_files(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenant_data_1 = """
        - tenant:
            name: local
            source:
              gerrit:
                config-projects:
                  - config
        """
        tenant_data_2 = """
        - tenant:
            name: ansible-network
            max-nodes-per-job: 5
            source:
              gerrit:
                config-projects:
                  - config
              github:
                untrusted-projects:
                  - repo1
                  - repo2
                  - include: []
                    projects:
                      - repo3
        """
        tenant_data_3 = """
        - tenant:
            name: local2
            max-nodes-per-job: 5
        """
        final_tenants = {}
        projects_list = {}
        tenant1 = yaml.safe_load(tenant_data_1)[0]
        tenant2 = yaml.safe_load(tenant_data_2)[0]
        tenant3 = yaml.safe_load(tenant_data_3)[0]
        ztl.merge_tenant_from_data(
            final_tenants, tenant1, '/t1', 'local', projects_list)
        ztl.merge_tenant_from_data(
            final_tenants, tenant2, '/t2', 'ansible-network', projects_list)
        ztl.merge_tenant_from_data(
            final_tenants, tenant3, '/t3', 'local2', projects_list)
        projects_list_expected = {
            'local': ['config'],
            'ansible-network': ['config', 'repo1', 'repo2', 'repo3']}
        self.assertItemsEqual(projects_list['local'],
                              projects_list_expected['local'])
        self.assertItemsEqual(projects_list['ansible-network'],
                              projects_list_expected['ansible-network'])
        final_tenants = ztl.final_tenant_merge(final_tenants)
        expected_tenant_ansible_network = {
            'tenant': {
                'admin-rules': [
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN'
                ],
                'max-nodes-per-job': 5,
                'name': 'ansible-network',
                'source': {
                    'github': {
                        'untrusted-projects': [
                            'repo1',
                            'repo2',
                            {'include': [],
                             'projects': ['repo3']}
                            ]
                        },
                    'gerrit': {
                        'config-projects': ['config']
                        }
                    }
                }
            }
        expected_tenant_local = {
            'tenant': {
                'name': 'local',
                'admin-rules': [
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN'
                ],
                'source': {
                    'gerrit': {
                        'config-projects': ['config']
                        }
                    }
                }
            }
        # Verify tenants def w/o sources are supported
        expected_tenant_local2 = {
            'tenant': {
                'name': 'local2',
                'admin-rules': [
                    '__SF_DEFAULT_ADMIN',
                    '__SF_TENANT_ZUUL_ADMIN'
                ],
                'max-nodes-per-job': 5,
                }
            }
        self.assertDictEqual(
            [t for t in final_tenants if
             t['tenant']['name'] == 'ansible-network'][0],
            expected_tenant_ansible_network)
        self.assertDictEqual(
            [t for t in final_tenants if
             t['tenant']['name'] == 'local'][0],
            expected_tenant_local)
        self.assertDictEqual(
            [t for t in final_tenants if
             t['tenant']['name'] == 'local2'][0],
            expected_tenant_local2)
        self.assertEqual(len(final_tenants), 3)

    def test_merge_tenants_from_resources(self):
        ztl = ZuulTenantsLoad(utests=True)
        tenants = {
            'local': {
                'source': {
                    'gerrit': {
                        'config-projects': [
                            'config'
                        ]
                    }
                }
            },
            'ansible-network': {
                'source': {
                    'gerrit': {
                        'untrusted-projects': ['repo1', 'repo2'],
                        'config-projects': ['config'],
                    },
                },
                'max-nodes-per-job': 5}
        }
        orig_tenants = copy.deepcopy(tenants)
        projects_list = {
            'local': ['config'],
            'ansible-network': ['repo1', 'repo2', 'config']
        }
        local_resources = {
            'resources': {
                'connections': {
                    'gerrit': {}
                }
            }
        }
        resources = {
            'resources': {
                'projects': {
                    'project2': {
                        'connection': 'gerrit',
                        'options': ['zuul/skip'],
                        'source-repositories': [
                            {'repo4': {'zuul/include': []}}
                        ]
                    },
                    'project1': {
                        'connection': 'gerrit',
                        'source-repositories': [
                            {'repo3': {
                                'zuul/exclude-unprotected-branches': True,
                                'zuul/include': [],
                            }},
                            {'config2': {
                                'zuul/config-project': True,
                            }},
                            # Is already part of flat tenant definition then
                            # will be ignored
                            {'repo1': {
                                'zuul/include': [],
                            }},
                            {'repo4': {
                                'zuul/ignore': True,
                            }},
                            {'repo5': {
                                'private': True,
                            }},
                        ]
                    }
                }
            }
        }
        ztl.merge_tenant_from_resources(
            tenants, resources, 'ansible-network', projects_list,
            local_resources, 'gerrit')
        up = tenants['ansible-network']['source'][
            'gerrit']['untrusted-projects']
        cp = tenants['ansible-network']['source'][
            'gerrit']['config-projects']
        self.assertIn('config', cp)
        self.assertIn({'config2': {}}, cp)
        self.assertIn('repo1', up)
        self.assertIn('repo2', up)
        self.assertIn(
            {'repo3': {
                'exclude-unprotected-branches': True,
                'include': [],
            }}, up)
        self.assertNotIn('repo4', up)
        self.assertNotIn('repo5', up)

        local_resources = {
            'resources': {
                'connections': {
                    'gerrit2': {}
                }
            }
        }
        tenants = orig_tenants
        with self.assertRaises(RuntimeError) as ctx:
            ztl.merge_tenant_from_resources(
                tenants, resources, 'ansible-network', {},
                local_resources, 'gerrit')
        self.assertEqual(str(ctx.exception), "gerrit is an unknown connection")

    def test_tenant_options_from_resources(self):
        tenants = {}
        ztl = ZuulTenantsLoad(utests=True)
        tenant_options = {
            "tenant-options": {
                "zuul/default-jobs-timeout": "3600",
                "zuul/allowed-reporter": "review.openstack.org"
            }
        }
        ztl.merge_tenant_from_resources(
            tenants, {}, "third-party-ci-tenant", {}, {}, 'gerrit',
            tenant_options)
        final_data = ztl.final_tenant_merge(tenants)
        self.assertIn("default-jobs-timeout", final_data[0]['tenant'])
        self.assertEqual(final_data[0]['tenant']['default-jobs-timeout'],
                         "3600")

    def test_tenant_add_extra_repo(self):
        tenants = {}
        projects_list = {}
        local_resources = {
            'resources': {
                'connections': {
                    'gerrit': {}
                }
            }
        }
        tenant_resources = {
            'resources': {
                'projects': {
                    'project1': {
                        'tenant': 'tenant1',
                        'source-repositories': [
                            {'repo1': {'zuul/include': ['jobs']}},
                            {'repo3': {}},
                        ]
                    }
                },
                'repos': {
                    'repo1': {},
                    'repo2': {},
                }
            }
        }
        ztl = ZuulTenantsLoad(utests=True)
        ztl.merge_tenant_from_resources(
            tenants, tenant_resources, "tenant1", projects_list,
            local_resources, 'gerrit')
        ztl.add_missing_repos(
            tenants, tenant_resources, "tenant1", projects_list,
            local_resources, 'gerrit')
        up = tenants['tenant1']['source']['gerrit']['untrusted-projects']
        self.assertIn({'repo1': {'include': ['jobs']}}, up)
        self.assertIn({'repo2': {'include': []}}, up)
        self.assertIn({'repo3': {}}, up)
        self.assertEqual(len(up), 3)


class NodepoolConfTests(TestCase):

    def test_load(self):
        nc = NodepoolConf(
            config_dir='managesf/tests/fixtures',
            hostname='nl01',
            catchall=True)
        ret = nc.merge()
        expected_ret = {
            'elements-dir': '/etc/opt/rh/rh-python35/nodepool/elements:/usr/share/sf-elements',  # noqa
            'images-dir': '/var/opt/rh/rh-python35/lib/nodepool/dib',
            'providers': [
                {
                    'name': 'default',
                    'image-name-format': '{image_name}-{timestamp}',
                    'rate': 10.0,
                    'clean-floating-ips': True,
                    'pools': [
                        {
                            'labels':
                            [
                                {
                                    'min-ram': 1024,
                                    'name': 'dib-centos-7',
                                    'diskimage': 'dib-centos-7'
                                }
                            ],
                            'name': 'main',
                            'max-servers': 5,
                            'networks': ['slave-net-name']
                        }
                    ],
                    'boot-timeout': 120,
                    'diskimages': [{'name': 'dib-centos-7'}],
                    'cloud': 'default'
                }
            ],
            'labels': [{'name': 'dib-centos-7', 'min-ready': 1}],
            'build-log-retention': 7,
            'zookeeper-servers': [
                {
                    'host': 'managesf.sftests.com',
                    'port': 2181
                }
            ],
            'webapp': {'port': 8006},
            'build-log-dir': '/var/www/nodepool-log/',
            'diskimages': [
                {
                    'username': 'zuul-worker',
                    'env-vars': {
                        'DIB_GRUB_TIMEOUT': '0',
                        'DIB_IMAGE_CACHE': '/var/cache/nodepool/dib_cache',
                        'DIB_CHECKSUM': '1',
                        'TMPDIR': '/var/cache/nodepool/dib_tmp',
                        'REQUESTS_CA_BUNDLE': ''
                    },
                    'elements': [
                        'centos-minimal',
                        'nodepool-minimal',
                        'zuul-worker-user'
                    ],
                    'name': 'dib-centos-7'
                }
            ]
        }
        self.assertEqual(ret, yaml.safe_dump(expected_ret,
                                             default_flow_style=False))

        # nl02 should only have the extra provider
        old_provider = expected_ret['providers'][0]
        expected_ret['providers'] = [{
            'name': 'k1s',
            'driver': 'openshift',
        }]
        nc = NodepoolConf(
            config_dir='managesf/tests/fixtures', hostname='nl02')
        ret = nc.merge()
        self.assertEqual(ret, yaml.safe_dump(expected_ret,
                                             default_flow_style=False))

        # builder host should have all the providers
        expected_ret['providers'].insert(0, old_provider)
        nc = NodepoolConf(
            config_dir='managesf/tests/fixtures',
            hostname='nb01',
            builder=True)
        ret = nc.merge()
        self.assertEqual(ret, yaml.safe_dump(expected_ret,
                                             default_flow_style=False))


class CauthConfTests(TestCase):

    def test_load(self):
        rpc = CauthConf(utests=True)
        resources = {
            'resources': {
                'tenants': {
                    'local': {
                        'url': 'https://sftests.com/manage',
                        'default-connection': 'gerrit',
                    },
                    'tenant2': {
                        'url': 'https://sftests.com/manage',
                    }
                },
                'connections': {
                    'gerrit': {
                        'base-url': 'https://sftests.com/r',
                        'type': 'gerrit'
                    }
                },
                'projects': {
                    'project1': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo1': {}},
                        ]
                    },
                    'project2': {
                        'tenant': 'tenant2',
                        'source-repositories': [
                            {'repo3': {}},
                            {'repo4': {
                                'connection': 'gerrit'
                            }},
                            {'repo5': {
                                'private': True
                            }},
                        ]
                    },
                    'project3': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo6': {}},
                            {'repo7': {
                                'repoxplorer/skip': True,
                            }},
                        ]
                    },
                    'project4': {
                        'tenant': 'local',
                        'options': [
                            'repoxplorer/skip',
                        ],
                        'source-repositories': [
                            {'repo7': {}},
                        ]
                    },
                },
                'repos': {
                    'repo2': {},
                    'repo5': {},
                },
                'groups': {
                    'group1': {
                        'members': [
                            'admin@sftests.com'
                        ]
                    }
                }
            }
        }
        rpc.main_resources = resources
        ret = yaml.safe_load(rpc.start())
        expected_ret = {
            'groups': {
                'group1': {
                    'description': '',
                    'members': [
                        'admin@sftests.com',
                    ]
                }
            },
        }
        self.assertDictEqual(ret, expected_ret)


class HoundConfTests(TestCase):

    def test_load(self):
        rpc = HoundConf(utests=True)
        self.maxDiff = None
        resources = {
            'resources': {
                'tenants': {
                    'local': {
                        'url': 'https://sftests.com/manage',
                        'default-connection': 'gerrit',
                    },
                    'tenant2': {
                        'url': 'https://sftests.com/manage',
                    }
                },
                'connections': {
                    'gerrit': {
                        'base-url': 'https://sftests.com/r',
                        'type': 'gerrit'
                    },
                    'github': {
                        'base-url': 'https://github.com/',
                        'type': 'github'
                    },
                    'pagure': {
                        'base-url': 'https://pagure.io/',
                        'type': 'pagure'
                    },
                    'gitlab': {
                        'base-url': 'https://gitlab.com/',
                        'type': 'gitlab'
                    },
                    'gerrithub': {
                        'base-url': 'https://review.gerrithub.io/',
                        'type': 'gerrit'
                    }
                },
                'projects': {
                    'project1': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo1': {}},
                        ]
                    },
                    'project2': {
                        'tenant': 'tenant2',
                        'source-repositories': [
                            {'repo3': {}},
                            {'repo4': {
                                'connection': 'gerrit'
                            }},
                            {'repo5': {
                                'private': True
                            }},
                        ]
                    },
                    'project3': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo6': {}},
                            {'repo7': {
                                'hound/skip': True,
                            }},
                        ]
                    },
                    'project4': {
                        'tenant': 'local',
                        'options': [
                            'hound/skip',
                        ],
                        'source-repositories': [
                            {'repo7': {}},
                        ]
                    },
                    'project5': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo8': {
                                'connection': 'github'
                            }},
                            {'repo9': {
                                'connection': 'gerrithub'
                            }},
                        ]
                    },
                    'project6': {
                        'tenant': 'local',
                        'source-repositories': [
                            {'repo10': {
                                'connection': 'pagure',
                                'default-branch': 'main'
                            }},
                            {'repo11': {
                                'connection': 'gitlab'
                            }},
                        ]
                    },
                },
                'repos': {
                    'repo2': {},
                    'repo5': {},
                },
            }
        }
        rpc.main_resources = resources
        ret = yaml.safe_load(rpc.start())
        expected_ret = {
            'repos': {
                'repo2': {
                    'ms-between-poll': 43200000,
                    'url': 'https://sftests.com/r/repo2',
                    'vcs-config': {
                        'ref': 'master'
                     },
                    'url-pattern': {
                        'base-url': (
                            'https://sftests.com/r/plugins/gitiles/repo2'
                            '/+/refs/heads/master/{path}{anchor}'),
                        'anchor': '#{line}'
                    }
                },
                'repo1': {
                    'ms-between-poll': 43200000,
                    'url': 'https://sftests.com/r/repo1',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'base-url': (
                            'https://sftests.com/r/plugins/gitiles/repo1'
                            '/+/refs/heads/master/{path}{anchor}'),
                        'anchor': '#{line}'
                    }
                },
                'repo6': {
                    'ms-between-poll': 43200000,
                    'url': 'https://sftests.com/r/repo6',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'base-url': (
                            'https://sftests.com/r/plugins/gitiles/repo6'
                            '/+/refs/heads/master/{path}{anchor}'),
                        'anchor': '#{line}'
                    }
                },
                'repo4': {
                    'ms-between-poll': 43200000,
                    'url': 'https://sftests.com/r/repo4',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'base-url': (
                            'https://sftests.com/r/plugins/gitiles/repo4'
                            '/+/refs/heads/master/{path}{anchor}'),
                        'anchor': '#{line}'
                    }
                },
                'repo8': {
                    'ms-between-poll': 43200000,
                    'url': 'http://github.com/repo8',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'anchor': '#L{line}',
                        'base-url': (
                            'http://github.com/repo8/blob/master/'
                            '{path}{anchor}')
                    }
                },
                'repo9': {
                    'ms-between-poll': 43200000,
                    'url': 'https://review.gerrithub.io/repo9',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'base-url': (
                            'http://github.com/repo9/blob/master/'
                            '{path}{anchor}'),
                        'anchor': '#L{line}'
                    }
                },
                'repo10': {
                    'ms-between-poll': 43200000,
                    'url': 'https://pagure.io/repo10',
                    'vcs-config': {
                        'ref': 'main'
                    },
                    'url-pattern': {
                        'base-url': (
                            'https://pagure.io/repo10/blob/main/f/'
                            '{path}{anchor}'),
                        'anchor': '#_{line}'
                    }
                },
                'repo11': {
                    'ms-between-poll': 43200000,
                    'url': 'https://gitlab.com/repo11',
                    'vcs-config': {
                        'ref': 'master'
                    },
                    'url-pattern': {
                        'base-url': (
                            'https://gitlab.com/repo11/-/blob/master/'
                            '{path}{anchor}'),
                        'anchor': '#L{line}'
                    }
                },
            },
            'dbpath': '/var/lib/hound/data',
            'max-concurrent-indexers': 2
        }
        self.assertDictEqual(ret, expected_ret)



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from pecan import conf
from pecan.rest import RestController
from pecan import request, response
from stevedore import driver

from managesf.controllers import introspection
from managesf.services import base
from managesf import policy

from managesf import DEFAULT_SERVICES
from managesf.controllers.api.v2 import resources as v2_resources
from managesf.controllers.api.v2 import configurations as v2_configurations


logger = logging.getLogger(__name__)

LOGERRORMSG = "Unable to process client request, failed with "\
              "unhandled error: %s"
CLIENTERRORMSG = "Unable to process your request, failed with "\
                 "unhandled error (server side): %s"

# TODO: add detail (detail arg or abort function) for all abort calls.


# instanciate service plugins
SF_SERVICES = []
SERVICES = {}


def load_services():
    try:
        if conf.services:
            services = conf.services
        else:
            services = DEFAULT_SERVICES
            msg = 'No service configured, loading: %s' % DEFAULT_SERVICES
            logger.info(msg)
    except AttributeError:
        services = DEFAULT_SERVICES
        msg = 'Obsolete conf file, loading default: %s' % DEFAULT_SERVICES
        logger.info(msg)

    for service in services:
        try:
            plugin = driver.DriverManager(namespace='managesf.service',
                                          name=service,
                                          invoke_on_load=True,
                                          invoke_args=(conf,)).driver
            SF_SERVICES.append(plugin)
            SERVICES[service] = plugin
            logger.info('%s plugin loaded successfully' % service)
        except Exception as e:
            logger.error('Could not load service %s: %s' % (service, e))


def is_admin(user):
    return base.RoleManager.is_admin(user)


def report_unhandled_error(exp):
    logger.exception(LOGERRORMSG % str(exp))
    response.status = 500
    return CLIENTERRORMSG % str(exp)

# TODO move to utils and use resources rather than gerrit for groups


def authorize(rule_name, target):
    if not request.remote_user:
        request.remote_user = request.headers.get('X-Remote-User')
    credentials = {'username': request.remote_user, 'groups': []}
    # OpenID Connect authentication
    if request.headers.get("OIDC_CLAIM_groups", None) is not None:
        for group in request.headers.get('OIDC_CLAIM_groups').split(','):
            # it seems like keycloak prefixes groups with /, remove it
            if group.startswith('/'):
                credentials['groups'].append(group[1:])
            else:
                credentials['groups'].append(group)
    # gerrit based
    else:
        if request.remote_user:
            code_reviews = [s for s in SF_SERVICES
                            if isinstance(s, base.BaseCodeReviewServicePlugin)]
            if code_reviews:
                user_groups = code_reviews[0].project.get_user_groups(
                    request.remote_user)
                credentials['groups'] = [grp['name'] for grp in user_groups]
    return policy.authorize(rule_name, target, credentials)


load_services()


# API v2 - will be in its own file hierarchy once dependencies are sorted out


class V2Controller(object):
    # Mimic api v1 and replace endpoints incrementally
    about = introspection.IntrospectionController()
    resources = v2_resources.ResourcesRootController()
    configurations = v2_configurations.ConfigurationController()


class RootController(object):
    def __init__(self, *args, **kwargs):
        try:
            # just try to get the api config
            _ = conf.api.v2  # noQA
            self.v2 = V2Controller()
        except AttributeError:
            # TODO have a generic blank REST controller that returns
            # 'Not Implemented' error code
            logger.info('API v2 is not configured, skipping endpoint.')
            self.v2 = RestController()

        self.about = introspection.IntrospectionController()



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil

from unittest import TestCase
from webtest import TestApp
from pecan import load_app


from managesf.tests import dummy_conf

# plugins imports
# TODO: should be done dynamically depending on what plugins we want


def raiseexc(*args, **kwargs):
    raise Exception('FakeExcMsg')


class FunctionalTest(TestCase):
    def setUp(self):
        self.to_delete = []
        c = dummy_conf()
        self.config = {'services': c.services,
                       'gerrit': c.gerrit,
                       'app': c.app,
                       'admin': c.admin,
                       'sqlalchemy': c.sqlalchemy,
                       'managesf': c.managesf,
                       'policy': c.policy,
                       'resources': c.resources,
                       'api': c.api, }
        # deactivate loggin that polute test output
        # even nologcapture option of nose effetcs
        # 'logging': c.logging}
        self.app = TestApp(load_app(self.config))

    def tearDown(self):
        # Remove the sqlite db
        os.unlink(self.config['sqlalchemy']['url'][len('sqlite:///'):])
        for path in self.to_delete:
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)


class TestManageSFAPIv2(FunctionalTest):
    def test_get_api_endpoint(self):
        response = self.app.get('/v2/about/').json
        self.assertEqual('managesf',
                         response['service']['name'])
        self.assertEqual(set(self.config['services']),
                         set(response['service']['services']))


class TestManageSFIntrospectionController(FunctionalTest):

    def test_instrospection(self):
        response = self.app.get('/about/').json
        self.assertEqual('managesf',
                         response['service']['name'])
        self.assertEqual(set(self.config['services']),
                         set(response['service']['services']))


# def project_get(*args, **kwargs):
#     if kwargs.get('by_user'):
#         return ['p1', ]
#     return ['p0', 'p1']



#
# Copyright (C) 2017 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import abc
import six
import logging

from stevedore import driver

from pecan import conf

# TODO move exceptions somewhere more generic
from managesf.services import exceptions as exc


logger = logging.getLogger(__name__)


def load_manager(namespace, service):
    logger.info('loading %s:%s manager' % (namespace, service))
    # hard-coded Dummy service for testing. What could go wrong?
    if service == 'DummyService':
        return None
    try:
        manager = driver.DriverManager(namespace=namespace,
                                       name=service,
                                       invoke_on_load=True,
                                       invoke_args=(conf,)).driver
        logger.info('%s:%s manager loaded successfully' % (namespace,
                                                           service))
        return manager
    except Exception as e:
        msg = 'Could not load manager %s:%s: %s' % (namespace,
                                                    service, e)
        logger.error(msg)
        return None


@six.add_metaclass(abc.ABCMeta)
class BaseCRUDManager(object):

    def __init__(self, *args, **kwargs):
        self.ordering_options = []
        self._logger = logging.getLogger(
            'managesf.api.v2.%s' % self.__class__.__name__)

    @abc.abstractmethod
    def get(self, **kwargs):
        """get one or many items"""
        raise NotImplementedError

    def create(self, **kwargs):
        """create an item"""
        raise NotImplementedError

    def update(self, **kwargs):
        """update an item"""
        raise NotImplementedError

    def delete(self, **kwargs):
        """delete an item"""
        raise NotImplementedError


@six.add_metaclass(abc.ABCMeta)
class BaseService(object):
    """Base plugin for a service that can be managed by Software Factory (v2).
    """

    _config_section = "base"
    service_name = "base service"

    def __init__(self, conf):
        self._full_conf = conf
        try:
            self.configure_service(conf)
        except AttributeError:
            raise Exception(repr(conf))

    def configure_service(self, conf):
        try:
            self.conf = getattr(conf, self._config_section, None)
        except KeyError:
            msg = ("The %s service is not available" % self._config_section)
            raise exc.ServiceNotAvailableError(msg)
        if not self.conf:
            msg = ("The %s service is not available" % self._config_section)
            raise exc.ServiceNotAvailableError(msg)



# Copyright (C) 2018 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import git
import json
import logging
import os
import requests
import yaml

from pecan import conf
from pecan import expose

from managesf.controllers.api.v2 import base
from managesf.model.yamlbkd.engine import SFResourceBackendEngine


class BaseConfigurationController(base.APIv2RestController):
    def __init__(self, engine):
        self.engine = engine


class ZuulConfigurationController(BaseConfigurationController):
    @expose()
    def get(self, **kwargs):
        return ZuulTenantsLoad(
            engine=self.engine,
            default_tenant_name=conf.resources.get('tenant_name', 'local')
            ).start()


class NodepoolConfigurationController(base.APIv2RestController):
    @expose()
    def get(self, **kwargs):
        return NodepoolConf(via_web=True).start()


class HoundConfigurationController(BaseConfigurationController):
    @expose()
    def get(self, **kwargs):
        return HoundConf(
            engine=self.engine,
            default_tenant_name=conf.resources.get('tenant_name', 'local')
            ).start()


class CauthConfigurationController(BaseConfigurationController):
    @expose()
    def get(self, **kwargs):
        return CauthConf(
            engine=self.engine,
            default_tenant_name=conf.resources.get('tenant_name', 'local')
            ).start()


class ConfigurationController:
    def __init__(self):
        self.engine = SFResourceBackendEngine(
            os.path.join(conf.resources['workdir'], 'read'),
            conf.resources['subdir'])

        self.zuul = ZuulConfigurationController(self.engine)
        self.nodepool = NodepoolConfigurationController()
        self.hound = HoundConfigurationController(self.engine)
        self.cauth = CauthConfigurationController(self.engine)


def get_resources(url):
    """Get resources and config location from tenant deployment."""
    ret = requests.get(url, verify='/etc/ssl/certs/ca-bundle.crt')
    res = ret.json()
    if res.get('error_description'):
        raise RuntimeError("%s: failed %s" % (url, res))
    return res


class ZuulTenantsLoad:
    log = logging.getLogger("managesf.ZuulTenantsLoad")

    def __init__(self, engine=None, utests=False,
                 cache_dir="/var/lib/managesf/git",
                 default_tenant_name="local",
                 config_dir=None,
                 gateway_url=None,
                 tenant=None,
                 master_sf_url=None):
        self.main_resources = {}
        self.cache_dir = cache_dir
        self.default_tenant_name = default_tenant_name
        self.tenant_resources = None
        self.gateway_url = gateway_url
        self.utests = utests
        self.default_auth_rule_name = "__SF_DEFAULT_ADMIN"
        self.default_auth_rule = {
            'name': self.default_auth_rule_name,
            'conditions': [
                {'username': 'admin'},
                {'roles': 'zuul_admin'}
            ]
        }
        self.default_tenant_rule_name = "__SF_TENANT_ZUUL_ADMIN"
        self.default_tenant_rule = {
            "name": self.default_tenant_rule_name,
            'conditions': [
                {'roles': '{tenant.name}_zuul_admin'}
            ]
        }
        if self.utests:
            # Skip this for unittests
            return
        if not os.path.isdir(self.cache_dir):
            os.makedirs(self.cache_dir)
        if config_dir and gateway_url:
            # Direct use of the resources engine to read
            # resources from a config repo copy.
            # Main usage is by the config-check job
            eng = SFResourceBackendEngine(
                os.path.join(self.cache_dir, 'validate'), 'resources')
            resources = eng.get(
               'file://%s' % config_dir,
               'master', '%s/manage' % gateway_url.rstrip('/'))
            if tenant and master_sf_url:
                self.main_resources = get_resources(
                    "%s/manage/v2/resources" % master_sf_url)
                self.tenant_resources = resources
            else:
                self.main_resources = resources
        elif engine is None:
            # From cli uses api instead
            self.main_resources = get_resources(
                "http://localhost:20001/v2/resources")
        else:
            self.main_resources = engine.get(
                conf.resources['master_repo'], 'master')

    # Legacy zuul flat files handling
    def fetch_git_repo(self, tenant, url, store, ssl_verify=True):
        """Get config repository from tenant deployment"""
        tenant_config_project_path = os.path.join(
            store, tenant, os.path.basename(url))
        if not os.path.isdir(tenant_config_project_path):
            os.makedirs(tenant_config_project_path)
        if not os.path.isdir(os.path.join(tenant_config_project_path, '.git')):
            # Clone this is the first time
            git.Repo.clone_from(
                url, tenant_config_project_path, branch='master',
                config='http.sslVerify=%s' % bool(int(ssl_verify)))
        else:
            # Refresh the repository
            repo = git.Repo(tenant_config_project_path)
            repo.remotes.origin.pull()
        return tenant_config_project_path

    def discover_yaml_files(self, tenants_dir):
        """Fetch legacy zuul yaml file from config repository"""
        paths = []
        for root, dirs, files in os.walk(tenants_dir, topdown=True):
            paths.extend([os.path.join(root, path) for path in files])
        # Keeps only .yaml files
        paths = filter(
            lambda x: x.endswith('.yaml') or x.endswith('.yml'), paths)
        return paths

    def sanitize_projects_list(self, projects_list):
        """Convert flat files projects list to strings list"""
        for tenant, tenant_projects_list in projects_list.items():
            sanitized_project_names = []
            for kp in tenant_projects_list:
                if isinstance(kp, dict) and 'projects' in kp.keys():
                    sanitized_project_names.extend(kp['projects'])
                elif isinstance(kp, dict):
                    sanitized_project_names.append(list(kp)[0])
                else:
                    sanitized_project_names.append(kp)
            projects_list[tenant] = sanitized_project_names

    def merge_source(
            self, tenant_conf, sources, path, tenant_name, projects_list):
        """Merge sources that may be defined in different objects"""
        tenant_sources = tenant_conf.setdefault("source", {})
        for source in sources.keys():
            source_conf = tenant_sources.setdefault(source, {})
            for project_type in sources[source]:
                projects = source_conf.setdefault(project_type, [])
                for project in sources[source][project_type]:
                    if project in projects:
                        raise RuntimeError(
                            "%s: define existing project %s for tenant %s"
                            % (path, project, tenant_name))
                    projects_list.setdefault(tenant_name, []).append(project)
                    projects.append(project)

    def merge_tenant_from_data(
            self, tenants, tenant, path, tenant_name, projects_list):
        """Load legacy zuul data object into the tenants object"""
        tenant = tenant.get('tenant')
        if tenant['name'] != tenant_name:
            return
        tenant_conf = tenants.setdefault(
            tenant['name'], {})
        for name, value in tenant.items():
            if name != "name":
                self.log.debug("  -> %s: %s" % (name, value))
            if name == "source":
                # Merge source lists
                self.merge_source(
                    tenant_conf, value, path, tenant_name, projects_list)
            elif name != "name":
                # Set tenant option
                if name in tenant_conf:
                    raise RuntimeError(
                        "%s: define multiple %s for tenant %s" % (
                            path, tenant["name"], name))
                tenant_conf[name] = value
        self.sanitize_projects_list(projects_list)

    def merge_auth_rule_from_data(self, auth_rules, auth_rule):
        auth_rule = auth_rule.get('authorization-rule')
        rule_name = auth_rule.get('name')
        auth_rules[rule_name] = {
            'conditions': auth_rule['conditions']
        }

    def merge_tenant_from_files(
            self, tenants, auth_rules, tenants_conf_files,
            tenant_name, projects_list):
        """Load legacy zuul yaml file into the tenants object"""
        self.log.debug('Merge tenant files for tenant %s' % tenant_name)
        for path in tenants_conf_files:
            data = yaml.safe_load(open(path))
            if not data:
                continue
            self.log.debug(" Merge tenant from data (%s)" % path)
            for data_block in data:
                if not isinstance(data_block, dict):
                    raise RuntimeError(
                        "%s: invalid Zuul configuration block: %s" % (
                            path, data_block))
                if data_block.get('tenant'):
                    tenant = data_block
                    self.merge_tenant_from_data(
                        tenants, tenant, path, tenant_name, projects_list)
                elif data_block.get('authorization-rule'):
                    auth_rule = data_block
                    self.merge_auth_rule_from_data(
                        auth_rules, auth_rule)
                else:
                    raise RuntimeError(
                        "%s: invalid Zuul configuration block: %s" % (
                            path, data_block))

    # Zuul configuration loading directly from resources
    def merge_tenant_from_resources(
            self, tenants, tenant_resources, tenant_name, projects_list,
            local_resources, default_conn, tenant_conf={}):
        self.log.debug('Merge resources for tenant %s' % tenant_name)
        # Set zuul-tenant-option
        tenant_options = tenant_conf.get("tenant-options", {})
        for name, value in tenant_options.items():
            if name.startswith('zuul/'):
                tenants.setdefault(
                    tenant_name, {})[name.replace('zuul/', '')] = value

        for project_name, project in tenant_resources.get(
                'resources', {}).get('projects', {}).items():
            if not project.get('tenant'):
                if tenant_resources == self.main_resources and \
                      tenant_name != self.default_tenant_name:
                    # This is a tenant hosted on the master instance,
                    # we only match project with an explicit tenant name
                    continue
                project['tenant'] = tenant_name
            else:
                if project['tenant'] != tenant_name:
                    continue
            for sr in project['source-repositories']:
                sr_name = list(sr)[0]
                if (sr[sr_name].get('zuul/skip') is True or
                        'zuul/skip' in project.get('options', [])):
                    continue
                if (tenant_name in projects_list and
                        sr_name in projects_list[tenant_name]):
                    # already defined in flat zuul tenant file
                    continue
                else:
                    projects_list.setdefault(tenant_name, []).append(sr_name)
                    if sr[sr_name].get('zuul/ignore') is True:
                        continue
                    if sr[sr_name].get('private') is True:
                        continue
                source = (sr[sr_name].get('connection') or
                          project.get('connection', default_conn))
                if source not in local_resources['resources']['connections']:
                    raise RuntimeError("%s is an unknown connection" % source)
                _project = {sr_name: {}}
                sr_type = 'untrusted-projects'
                if sr[sr_name].get('zuul/config-project') is True:
                    sr_type = 'config-projects'
                    del sr[sr_name]['zuul/config-project']
                for option, value in sr[sr_name].items():
                    if option.startswith('zuul/'):
                        _project[sr_name][option.replace('zuul/', '')] = value
                self.log.debug(" -> Adding %s to %s" % (_project, source))
                tenants.setdefault(
                    tenant_name, {}).setdefault(
                        'source', {}).setdefault(
                            source, {}).setdefault(
                                sr_type, []).append(_project)

    def add_missing_repos(self, tenants, tenant_resources, tenant_name,
                          projects_list, local_resources, default_conn):
        self.log.debug('Merge missing repos for tenant %s' % tenant_name)
        tenant_repos = tenant_resources.get(
            'resources', {}).get('repos', {}).items()
        r_type = 'untrusted-projects'
        for repo_name, repo in tenant_repos:
            if not [True for v in projects_list.values() if repo_name in v]:
                _project = {repo_name: {'include': []}}
                self.log.debug("-> Adding %s to %s" % (_project, default_conn))
                tenants.setdefault(
                    tenant_name, {}).setdefault(
                        'source', {}).setdefault(
                            default_conn, {}).setdefault(
                                r_type, []).append(_project)
            projects_list.setdefault(tenant_name, []).append(repo_name)

    def final_tenant_merge(self, tenants):
        final_data = []
        for tenant, tenant_conf in sorted(tenants.items()):
            data = {'tenant': {'name': tenant}}
            data['tenant'].update(tenant_conf)
            # if the default admin-rules are already set, do not add them
            admin_rules = data['tenant'].get('admin-rules', [])
            for rule_name in [
                self.default_auth_rule_name,
                self.default_tenant_rule_name,
            ]:
                if rule_name not in admin_rules:
                    admin_rules.append(rule_name)
            data['tenant']['admin-rules'] = admin_rules
            final_data.append(data)
        return final_data

    def merge_auth_rules(self, auth_rules):
        final = [
            {'authorization-rule': self.default_auth_rule},
            {'authorization-rule': self.default_tenant_rule},
        ]
        for rule_name, rule in sorted(auth_rules.items()):
            if rule_name in [
                self.default_auth_rule_name,
                self.default_tenant_rule_name,
            ]:
                # we do not allow overriding this rule
                continue
            data = {'authorization-rule': {'name': rule_name}}
            data['authorization-rule'].update(rule)
            final.append(data)
        return final

    def start(self):
        """Generate a zuul main.yaml from managesf resources and flat files"""

        # tenants is the main structure to be converted into zuul main.yaml
        tenants = {}
        # projects_list is the list of projects used to check for conflicts
        projects_list = {}
        tenant_resources_cache = {}

        auth_rules = {}

        for tenant_name, tenant_conf in self.main_resources.get(
                "resources", {}).get("tenants", {}).items():

            # When a tenant_resources already exists
            # It is only set if we loaded tenant resources from
            # a config repo copy (eg. config-check)
            if self.tenant_resources:
                # We only proceed when the args gateway_url
                # match the tenant url
                if tenant_conf['url'] != '%s/manage' % (
                        self.gateway_url.rstrip('/')):
                    continue

            self.log.debug(
                "--[ Processing %s - %s" % (tenant_name, tenant_conf))

            # First we look for the tenant resources
            if tenant_name != self.default_tenant_name and \
               tenant_conf["url"] != self.main_resources["public-url"]:
                # check for v2
                tenant_url = tenant_conf['url'].rstrip('/')
                if not tenant_url.endswith('/v2'):
                    tenant_url = os.path.join(tenant_url, "v2")
                url = os.path.join(tenant_url, 'resources')
                if self.tenant_resources:
                    self.log.debug("%s: loading resources from workspace",
                                   tenant_name)
                    tenant_resources = self.tenant_resources
                else:
                    self.log.debug("%s: loading resources %s",
                                   tenant_name, url)
                    tenant_resources = get_resources(url)
            else:
                tenant_resources = self.main_resources
                # Fallback to default_tenant_name tenant default connection
                if not tenant_conf.get("default-connection"):
                    tenant_conf["default-connection"] = self.main_resources[
                        "resources"]["tenants"][self.default_tenant_name][
                            "default-connection"]

            tenant_resources_cache[tenant_name] = tenant_resources

            if not self.utests:
                # Then we pull tenant config repository for legacy zuul
                # flat files
                path = self.fetch_git_repo(
                    tenant_name, tenant_resources["config-repo"],
                    self.cache_dir)
                tenants_dir = os.path.join(path, 'zuul')
                if not os.path.isdir(tenants_dir):
                    continue
                tenants_conf_files = self.discover_yaml_files(tenants_dir)
                # And we load flat files
                self.merge_tenant_from_files(
                    tenants, auth_rules, tenants_conf_files,
                    tenant_name, projects_list)

            # We load project from the resources
            default_conn = tenant_conf["default-connection"]
            self.merge_tenant_from_resources(
                tenants, tenant_resources, tenant_name, projects_list,
                self.main_resources, default_conn, tenant_conf)

        for tenant_name, tenant_conf in self.main_resources.get(
                "resources", {}).get("tenants", {}).items():

            tenant_resources = tenant_resources_cache.get(tenant_name)
            if not tenant_resources:
                continue

            default_conn = tenant_conf["default-connection"]

            # Finally we add Repos not listed in sr with an include: [] to Zuul
            skip_missing_resources = False
            if tenant_conf["url"] == self.main_resources["public-url"]:
                if tenant_name != self.default_tenant_name:
                    # We only add local missing resources to the
                    # default_tenant_name tenant
                    skip_missing_resources = True
            # Check default_conn is a registered connection
            if default_conn not in self.main_resources[
                    'resources']['connections']:
                # We cannot add repos to Zuul if no valid connection for
                # that tenant
                self.log.debug(
                    "Skip adding missing repos. The tenant has an invalid"
                    " default connection: %s" % default_conn)
                continue
            if not skip_missing_resources:
                self.add_missing_repos(
                    tenants, tenant_resources, tenant_name, projects_list,
                    self.main_resources, default_conn)

            self.log.debug("]-- Finish processing %s" % tenant_name)

        final_tenant_data = self.final_tenant_merge(tenants)
        auth_rules_data = self.merge_auth_rules(auth_rules)
        final_data = auth_rules_data + final_tenant_data
        return yaml.safe_dump(final_data)


class NodepoolConf():
    log = logging.getLogger("managesf.NodepoolConf")

    def __init__(self, via_web=None,
                 cache_dir="/var/lib/managesf/git",
                 config_dir=None,
                 builder=False,
                 catchall=False,
                 hostname=None):
        self.cache_dir = cache_dir
        self.hostname = hostname
        self.builder = builder
        self.catchall = catchall

        if via_web:
            self.config_repo_path = self.fetch_git_repo(
                conf.resources['tenant_name'],
                conf.resources["master_repo"],
                self.cache_dir)
        else:
            self.config_repo_path = config_dir or '/root/config'

    def fetch_git_repo(self, tenant, url, store, ssl_verify=True):
        """Get config repository from tenant deployment"""
        tenant_config_project_path = os.path.join(
            store, tenant, os.path.basename(url))
        if not os.path.isdir(tenant_config_project_path):
            os.makedirs(tenant_config_project_path)
        if not os.path.isdir(os.path.join(tenant_config_project_path, '.git')):
            # Clone this is the first time
            git.Repo.clone_from(
                url, tenant_config_project_path, branch='master',
                depth=1, config='http.sslVerify=%s' % bool(int(ssl_verify)))
        else:
            # Refresh the repository
            repo = git.Repo(tenant_config_project_path)
            repo.remotes.origin.pull()
        return tenant_config_project_path

    def yaml_merge_load(self, nodepool_dir, _nodepool_conf):
        paths = []
        for file in filter(
                lambda x: x.endswith('.yaml') or x.endswith('.yml'),
                os.listdir(nodepool_dir)):
            paths.append(os.path.join(nodepool_dir, file))

        user = {}
        for path in paths:
            # Do not load base config
            if _nodepool_conf in path:
                continue
            data = yaml.safe_load(open(path))
            if not data:
                continue
            for key, value in data.items():
                user.setdefault(key, []).extend(value)
        return user

    def merge(self):
        nodepool_dir = '%s/nodepool' % self.config_repo_path
        _nodepool_conf = '%s/_nodepool.yaml' % nodepool_dir

        user = self.yaml_merge_load(nodepool_dir, _nodepool_conf)
        conf = yaml.safe_load(open(_nodepool_conf))

        # Builder service always catch all
        catchall = True if self.builder else self.catchall

        cache_dir = "/var/cache/nodepool"

        for dib in user.get('diskimages', []):
            self.log.debug('dibimage: %s' % dib)
            dib.setdefault('username', 'zuul-worker')
            envvars = dib.setdefault('env-vars', {})
            envvars['TMPDIR'] = "%s/dib_tmp" % cache_dir
            envvars['DIB_IMAGE_CACHE'] = "%s/dib_cache" % cache_dir
            envvars['DIB_GRUB_TIMEOUT'] = '0'
            envvars['DIB_CHECKSUM'] = '1'
            # Ensure host CA bundle doesn't interfer with dib
            envvars['REQUESTS_CA_BUNDLE'] = ''
            # Make sure env-vars are str
            for k, v in envvars.items():
                if not isinstance(v, str):
                    envvars[k] = str(v)

        if 'cron' in user:
            conf['cron'] = user['cron']
        conf['labels'] = user.get('labels', [])
        providers = user.get('providers', [])
        conf['diskimages'] = user.get('diskimages', [])
        for extra_labels in user.get('extra-labels', []):
            added = False
            for provider in providers:
                if provider['name'] != extra_labels['provider']:
                    continue
                if extra_labels.get('cloud-images'):
                    provider.setdefault('cloud-images', []).extend(
                        extra_labels['cloud-images'])
                for pool in provider.get('pools', []):
                    if pool['name'] == extra_labels['pool']:
                        pool['labels'].extend(extra_labels['labels'])
                        added = True
                        break
                if added:
                    break
            if not added:
                raise RuntimeError("%s: couldn't find provider" % extra_labels)

        # Ensure zuul-console-dir is removed from runc provider
        for provider in providers:
            if provider.get("zuul-console-dir"):
                provider.pop("zuul-console-dir")

        # Add providers
        conf['providers'] = []
        for provider in providers:
            if provider.get("launcher-host") is not None:
                if self.builder or provider["launcher-host"] == self.hostname:
                    conf['providers'].append(provider)
                provider.pop("launcher-host")
            elif catchall:
                conf['providers'].append(provider)

        self.log.debug('final conf: %s' % conf)
        return yaml.safe_dump(conf, default_flow_style=False)

    def start(self):
        """Generate nodepool.yaml file from managesf flat files"""
        return self.merge()


class HoundConf():
    log = logging.getLogger("managesf.HoundConf")

    def __init__(self, engine=None,
                 utests=False,
                 master_sf_url=None,
                 default_tenant_name='local'):
        self.default_tenant_name = default_tenant_name
        self.master_sf_url = master_sf_url
        self.privates = set()
        self.config = {
            "max-concurrent-indexers": 2,
            "dbpath": "/var/lib/hound/data",
            "repos": {},
        }
        if utests:
            # Skip this for unittests
            return
        if engine is None:
            # From cli uses api instead
            self.main_resources = get_resources(
                "http://localhost:20001/v2/resources")
        else:
            self.main_resources = engine.get(
                conf.resources['master_repo'], 'master')

    def compute_uri_gitweb(self, conn):
        conn_type = self.main_resources['resources'][
            'connections'][conn]['type']
        base_url = self.main_resources['resources'][
            'connections'][conn]['base-url'].rstrip('/')
        if conn_type == 'gerrit':
            uri = base_url + '/%(name)s'
            gitweb = (
                base_url +
                '/plugins/gitiles/%(name)s/+/' +
                'refs/heads/%(branch)s/{path}{anchor}')
            anchor = '#{line}'
            if 'https://review.gerrithub.io' in base_url:
                gitweb = (
                    'http://github.com/%(name)s/blob/%(branch)s/{path}{anchor}'
                )
                anchor = '#L{line}'
        if conn_type == 'github':
            uri = 'http://github.com/%(name)s'
            gitweb = (
                'http://github.com/%(name)s/blob/%(branch)s/{path}{anchor}')
            anchor = '#L{line}'
        if conn_type == 'pagure':
            uri = base_url + '/%(name)s'
            gitweb = (
                base_url + '/%(name)s/blob/%(branch)s/f/{path}{anchor}')
            anchor = '#_{line}'
        if conn_type == 'gitlab':
            uri = base_url + '/%(name)s'
            gitweb = base_url + '/%(name)s/-/blob/%(branch)s/{path}{anchor}'
            anchor = '#L{line}'

        return uri, gitweb, anchor

    def add_in_conf(self, repo, conn, branch):
        conn_type = self.main_resources['resources'][
            'connections'][conn]['type']
        if conn_type not in ['gerrit', 'github', 'pagure', 'gitlab']:
            return
        uri, gitweb, anchor = self.compute_uri_gitweb(conn)
        self.config["repos"][repo] = {
            "url": uri % {'name': repo},
            "ms-between-poll": int(12*60*60*1000),
            "vcs-config": {
                "ref": branch
            },
            "url-pattern": {
                "base-url": gitweb % {'name': repo, 'branch': branch},
                "anchor": anchor,
            }
        }

    def start(self):
        for project, data in self.main_resources.get(
                "resources", {}).get("projects", {}).items():
            for sr in data.get('source-repositories', []):
                repo_name = list(sr.keys())[0]
                if sr[repo_name].get('private') is True:
                    self.privates.add(repo_name)
                    continue
                if (sr[repo_name].get('hound/skip') is True or
                        'hound/skip' in data.get('options', [])):
                    self.privates.add(repo_name)
                    continue
                conn = sr[repo_name].get('connection')
                if not conn:
                    conn = data.get('connection')
                if not conn:
                    tenant_name = data.get('tenant', self.default_tenant_name)
                    tenant = self.main_resources['resources'][
                        'tenants'][tenant_name]
                    conn = tenant.get('default-connection')
                conn_defined = self.main_resources['resources'].get(
                    'connections', {}).get(conn)
                if conn_defined:
                    branch = sr[repo_name].get('default-branch', 'master')
                    self.add_in_conf(repo_name, conn, branch)
                    self.privates.add(repo_name)

        # Add repositories not associated to a project
        for repo, data in self.main_resources.get(
                "resources", {}).get("repos", {}).items():
            if repo in self.privates:
                continue
            tenant = self.main_resources['resources'][
                'tenants'][self.default_tenant_name]
            conn = tenant.get('default-connection')
            if conn_defined:
                branch = data.get('default-branch', 'master')
                self.add_in_conf(repo, conn, branch)

        return json.dumps(self.config, indent=True, sort_keys=True)


class CauthConf():
    log = logging.getLogger("managesf.CauthConf")

    def __init__(self, engine=None,
                 utests=False,
                 master_sf_url=None,
                 default_tenant_name='local'):
        self.default_tenant_name = default_tenant_name
        self.master_sf_url = master_sf_url
        self.repos_cache = set()
        self.default = {
            'groups': {},
        }
        if utests:
            # Skip this for unittests
            return
        if engine is None:
            # From cli uses api instead
            self.main_resources = get_resources(
                "http://localhost:20001/v2/resources")
        else:
            self.main_resources = engine.get(
                conf.resources['master_repo'], 'master')

    def start(self):
        # Add the groups
        for group, data in self.main_resources[
                'resources'].get('groups', {}).items():
            grp = {}
            grp['description'] = data.get('description', '')
            grp['members'] = data.get('members', [])
            # Only add groups with members
            if len(grp['members']) > 0:
                self.default['groups'][group] = grp

        return yaml.safe_dump(self.default)


def cli():
    import argparse

    p = argparse.ArgumentParser()
    p.add_argument("--output")
    p.add_argument("--config-dir")
    p.add_argument("--gateway-url")
    p.add_argument("--master-sf-url")
    p.add_argument("--extra-launcher", action='store_true', default=False,
                   help="This host is not the main nodepool-launcher")
    p.add_argument("--builder", action='store_true',
                   help="This host is a nodepool-builder")
    p.add_argument("--hostname", help="Get configuration for a dedicated host "
                   "(only for nodepool services)")
    p.add_argument("--tenant", action='store_true')
    p.add_argument(
        "--cache-dir", default="/var/lib/software-factory/git-configuration")
    p.add_argument("--debug", action="store_true")
    p.add_argument("--default-tenant-name", default="local")
    p.add_argument("service", choices=["zuul", "nodepool", "repoxplorer",
                                       "hound", "cauth"])
    args = p.parse_args()

    logging.basicConfig(
        format='%(asctime)s %(levelname)-5.5s %(name)s - %(message)s',
        level=logging.DEBUG if args.debug else logging.INFO)

    if args.service == "zuul":
        ztl = ZuulTenantsLoad(
            cache_dir=args.cache_dir,
            default_tenant_name=args.default_tenant_name,
            config_dir=args.config_dir,
            gateway_url=args.gateway_url,
            master_sf_url=args.master_sf_url,
            tenant=args.tenant)
        conf = ztl.start()

    if args.service == "nodepool":
        rpc = NodepoolConf(
            cache_dir=args.cache_dir,
            config_dir=args.config_dir,
            builder=args.builder,
            catchall=not args.extra_launcher,
            hostname=args.hostname)
        conf = rpc.start()

    if args.service == "hound":
        rpc = HoundConf(
            master_sf_url=args.master_sf_url,
            default_tenant_name=args.default_tenant_name)
        conf = rpc.start()

    if args.service == "cauth":
        rpc = CauthConf(
            master_sf_url=args.master_sf_url,
            default_tenant_name=args.default_tenant_name)
        conf = rpc.start()

    if args.output:
        open(args.output, "w").write(conf)
    else:
        print(conf)



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import tempfile

from pecan.configuration import conf_from_dict


class dummy_conf():
    services = ['SFGerrit',
                ]
    gerrit = {
        'url': 'http://gerrit.tests.dom',
        'password': 'admin_password',
        'host': 'gerrit.test.dom',
        'top_domain': 'tests.dom',
        'ssh_port': 2929,
        'sshkey_priv_path': tempfile.mkstemp()[1],
    }
    managesf = {
        'host': 'managesf.tests.dom',
        'sshkey_priv_path': '/tmp/id_rsa',
    }
    resources = {
        'workdir': tempfile.mkdtemp(),
        'subdir': 'resources',
        'master_repo': 'http://sftests.com/r/config',
        'public_url': 'http://sftests.com/manage',
    }
    admin = {
        'name': 'user1',
        'email': 'user1@tests.dom',
    }
    app = {
        'root': 'managesf.controllers.root.RootController',
        'template_path': os.path.join(os.path.dirname(__file__),
                                      '../templates'),
        'modules': ['managesf'],
        'debug': True,
    }
    sqlalchemy = {
        'url': 'sqlite:///%s' % tempfile.mkstemp()[1],
        'encoding': 'utf-8',
    }
    logging = {
        'loggers': {
            'root': {'level': 'INFO', 'handlers': ['console']},
            'managesf': {'level': 'DEBUG', 'handlers': ['console']},
            'py.warnings': {'handlers': ['console']},
            '__force_dict__': True},
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'}},
        'formatters': {
            'simple': {
                'format': ('%(asctime)s %(levelname)-5.5s [%(name)s]'
                           '[%(threadName)s] %(message)s')}}
    }
    policy = {}
    DummyService = {
        'dummy_opt1': 'value1',
        'dummy_opt2': 'value2'
    }
    api = conf_from_dict({
        'v2': {
            'resources': ['DummyService', ],
        },
    })



#
# Copyright (C) 2016 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -*- coding: utf-8 -*-


from oslo_policy import policy

from managesf.policies import base


BASE_POLICY_NAME = 'managesf.config'
POLICY_ROOT = BASE_POLICY_NAME + ':%s'

rules = [
    policy.RuleDefault(
        name=POLICY_ROOT % 'get',
        check_str=base.RULE_AUTHENTICATED_API),
]


def list_rules():
    return rules
